# -*- coding: utf-8 -*-
# sql.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Local blobs backend on SQLCipher
"""
import os
import binascii

from functools import partial
from io import BytesIO
from pysqlcipher import dbapi2
from twisted.internet import defer
from twisted.logger import Logger
from twisted.enterprise import adbapi

from leap.common.files import mkdir_p

from .. import sqlcipher
from .. import pragmas

logger = Logger()


class SyncStatus:
    SYNCED = 1
    PENDING_UPLOAD = 2
    PENDING_DOWNLOAD = 3
    FAILED_UPLOAD = 4
    FAILED_DOWNLOAD = 5
    PENDING_DELETE = 6
    LOCAL_ONLY = 7
    UNAVAILABLE_STATUSES = (3, 5)


class Priority:
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    URGENT = 4
    DEFAULT = 2


class SQLiteBlobBackend(object):

    concurrency_limit = 10

    def __init__(self, path, key=None, user=None):
        dbname = '%s_blobs.db' % (user or 'soledad')
        self.path = os.path.abspath(
            os.path.join(path, dbname))
        mkdir_p(os.path.dirname(self.path))
        if not key:
            raise ValueError('key cannot be None')
        backend = 'pysqlcipher.dbapi2'
        opts = sqlcipher.SQLCipherOptions(
            '/tmp/ignored', binascii.b2a_hex(key),
            is_raw_key=True, create=True)
        openfun = partial(pragmas.set_init_pragmas, opts=opts,
                          schema_func=_init_tables)

        self.dbpool = ConnectionPool(
            backend, self.path, check_same_thread=False, timeout=5,
            cp_openfun=openfun, cp_min=2, cp_max=2, cp_name='blob_pool')

    def close(self):
        from twisted._threads import AlreadyQuit
        try:
            self.dbpool.close()
        except AlreadyQuit:
            pass

    @defer.inlineCallbacks
    def put(self, blob_id, blob_fd, size=None,
            namespace=''):
        logger.info("Saving blob in local database...")
        insert = 'INSERT INTO blobs (blob_id, namespace, payload)'
        insert += ' VALUES (?, ?, zeroblob(?))'
        values = (blob_id, namespace, size)
        irow = yield self.dbpool.insertAndGetLastRowid(insert, values)
        yield self.dbpool.write_blob('blobs', 'payload', irow, blob_fd)
        logger.info("Finished saving blob in local database.")

    @defer.inlineCallbacks
    def get(self, blob_id, namespace=''):
        # TODO we can also stream the blob value using sqlite
        # incremental interface for blobs - and just return the raw fd instead
        select = 'SELECT payload FROM blobs WHERE blob_id = ? AND namespace= ?'
        values = (blob_id, namespace,)
        result = yield self.dbpool.runQuery(select, values)
        if result:
            defer.returnValue(BytesIO(str(result[0][0])))

    @defer.inlineCallbacks
    def get_sync_status(self, blob_id):
        select = 'SELECT sync_status, retries FROM sync_state WHERE blob_id= ?'
        result = yield self.dbpool.runQuery(select, (blob_id,))
        if result:
            defer.returnValue((result[0][0], result[0][1]))

    @defer.inlineCallbacks
    def get_sync_progress(self):
        query = 'SELECT sync_status, COUNT(sync_status) FROM sync_state'
        query += ' GROUP BY sync_status'

        def by_value(value):
            statuses = SyncStatus.__dict__.items()
            return filter(lambda x: x[1] == value, statuses)[0][0]
        result = yield self.dbpool.runQuery(query)
        if result:
            defer.returnValue(dict([(by_value(r[0]), r[1]) for r in result]))
        else:
            defer.returnValue([])

    @defer.inlineCallbacks
    def list(self, namespace=''):
        query = 'select blob_id from blobs where namespace = ?'
        values = (namespace,)
        result = yield self.dbpool.runQuery(query, values)
        if result:
            defer.returnValue([b_id[0] for b_id in result])
        else:
            defer.returnValue([])

    @defer.inlineCallbacks
    def list_status(self, sync_status, namespace=''):
        query = 'select blob_id from sync_state where sync_status = ?'
        query += 'AND namespace = ? ORDER BY priority DESC'
        values = (sync_status, namespace,)
        result = yield self.dbpool.runQuery(query, values)
        if result:
            defer.returnValue([b_id[0] for b_id in result])
        else:
            defer.returnValue([])

    @defer.inlineCallbacks
    def update_sync_status(self, blob_id, sync_status, namespace="",
                           priority=None):
        retries = 'SELECT retries from sync_state' \
                  ' WHERE blob_id="%s" and namespace="%s"' \
                  % (blob_id, namespace)
        if not priority:
            priority = 'SELECT priority FROM sync_state' \
                       ' WHERE blob_id="%s" AND namespace="%s"' \
                       % (blob_id, namespace)
        fields = 'blob_id, namespace, sync_status, retries, priority'
        markers = '?, ?, ?, (%s), COALESCE((%s), %s)' \
                  % (retries, priority, Priority.DEFAULT)
        values = [blob_id, namespace, sync_status]
        insert = 'INSERT or REPLACE INTO sync_state (%s) VALUES (%s)' \
                 % (fields, markers)
        yield self.dbpool.runOperation(insert, tuple(values))

    @defer.inlineCallbacks
    def get_priority(self, blob_id, namespace=""):
        query = 'SELECT priority FROM sync_state WHERE blob_id = ?'
        result = yield self.dbpool.runQuery(query, (blob_id,))
        if not result:
            defer.returnValue(None)
        priority = result.pop()[0]
        defer.returnValue(priority)

    @defer.inlineCallbacks
    def update_priority(self, blob_id, priority, namespace=""):
        old_priority = self.get_priority(blob_id, namespace=namespace)
        if not old_priority:
            logger.error("Can't update priority of %s: no sync status entry.")
            return
        if old_priority == priority:
            return
        update = 'UPDATE sync_state SET priority = ? WHERE blob_id = ?'
        values = (priority, blob_id,)
        yield self.dbpool.runOperation(update, values)

    def update_batch_sync_status(self, blob_id_list, sync_status,
                                 namespace=''):
        if not blob_id_list:
            return
        inserts = []
        values = ()
        for blob_id in blob_id_list:
            priority = 'SELECT priority FROM sync_state' \
                       ' WHERE blob_id="%s" AND namespace="%s"' \
                       % (blob_id, namespace)
            insert = '(?, ?, ?, COALESCE((%s), %s))' \
                     % (priority, Priority.DEFAULT)
            inserts.append(insert)
            values += (blob_id, namespace, sync_status)
        query = 'INSERT or REPLACE INTO sync_state'
        query += ' (blob_id, namespace, sync_status, priority) VALUES '
        query += ', '.join(inserts)
        return self.dbpool.runQuery(query, values)

    def increment_retries(self, blob_id):
        query = 'update sync_state set retries = retries + 1 where blob_id = ?'
        return self.dbpool.runQuery(query, (blob_id,))

    @defer.inlineCallbacks
    def list_namespaces(self):
        query = 'select namespace from blobs'
        result = yield self.dbpool.runQuery(query)
        if result:
            defer.returnValue([namespace[0] for namespace in result])
        else:
            defer.returnValue([])

    @defer.inlineCallbacks
    def get_size_list(self, blob_ids, namespace=''):
        query = 'SELECT BLOB_ID, LENGTH(PAYLOAD) FROM BLOBS WHERE BLOB_ID IN '
        query += ('(%s)' % ', '.join(['?' for _ in blob_ids]))
        query += ' AND NAMESPACE = ? ORDER BY BLOB_ID'
        values = tuple(blob_ids) + (namespace,)
        results = yield self.dbpool.runQuery(query, values)
        if results:
            results = dict([(result[0], result[1]) for result in results])
            # order results according to blob_ids parameter
            results = [(blob_id, results[blob_id]) for blob_id in blob_ids]
            defer.returnValue(results)
        else:
            defer.returnValue([])

    @defer.inlineCallbacks
    def exists(self, blob_id, namespace=''):
        query = 'SELECT blob_id from blobs WHERE blob_id = ? AND namespace= ?'
        result = yield self.dbpool.runQuery(query, (blob_id, namespace,))
        defer.returnValue(bool(len(result)))

    def delete(self, blob_id, namespace=''):
        query = 'DELETE FROM blobs WHERE blob_id = ? AND namespace = ?'
        return self.dbpool.runQuery(query, (blob_id, namespace,))

    def batch_delete(self, blob_id_list, namespace=''):
        query = 'DELETE FROM blobs WHERE blob_id IN '
        size = len(blob_id_list)
        query += ('(%s)' % ', '.join(['?' for _ in range(size)]))
        values = tuple(blob_id_list)
        return self.dbpool.runQuery(query, values)


def _init_tables(conn):
    # unified init for running under the same lock
    _init_blob_table(conn)
    _init_sync_table(conn)


def _init_sync_table(conn):
    maybe_create = """
        CREATE TABLE IF NOT EXISTS
        sync_state (
        blob_id PRIMARY KEY,
        namespace TEXT,
        sync_status INT default %s,
        priority INT default %d,
        retries INT default 0)"""
    default_status = SyncStatus.PENDING_UPLOAD
    default_priority = Priority.DEFAULT
    maybe_create %= (default_status, default_priority)
    conn.execute(maybe_create)


def _init_blob_table(conn):
    maybe_create = (
        "CREATE TABLE IF NOT EXISTS "
        "blobs ("
        "blob_id PRIMARY KEY, "
        "payload BLOB)")
    conn.execute(maybe_create)
    columns = [row[1] for row in conn.execute("pragma"
               " table_info(blobs)").fetchall()]
    if 'namespace' not in columns:
        # namespace migration
        conn.execute('ALTER TABLE blobs ADD COLUMN namespace TEXT')
    if 'sync_status' not in columns:
        # sync status migration
        default_status = SyncStatus.PENDING_UPLOAD
        sync_column = 'ALTER TABLE blobs ADD COLUMN sync_status INT default %s'
        sync_column %= default_status
        conn.execute(sync_column)
        conn.execute('ALTER TABLE blobs ADD COLUMN retries INT default 0')


class ConnectionPool(adbapi.ConnectionPool):

    timeout_retries = 20

    def insertAndGetLastRowid(self, *args, **kwargs):
        """
        Execute an SQL query and return the last rowid.

        See: https://sqlite.org/c3ref/last_insert_rowid.html
        """
        return self.runInteraction(
            self._insertAndGetLastRowid, *args, **kwargs)

    def _insertAndGetLastRowid(self, trans, *args, **kw):
        trans.execute(*args, **kw)
        return trans.lastrowid

    def blob(self, table, column, irow, flags):
        """
        Open a BLOB for incremental I/O.

        Return a handle to the BLOB that would be selected by:

          SELECT column FROM table WHERE rowid = irow;

        See: https://sqlite.org/c3ref/blob_open.html

        :param table: The table in which to lookup the blob.
        :type table: str
        :param column: The column where the BLOB is located.
        :type column: str
        :param rowid: The rowid of the BLOB.
        :type rowid: int
        :param flags: If zero, BLOB is opened for read-only. If non-zero,
                      BLOB is opened for RW.
        :type flags: int

        :return: A BLOB handle.
        :rtype: pysqlcipher.dbapi.Blob
        """
        return self.runInteraction(self._blob, table, column, irow, flags)

    def write_blob(self, table, column, irow, blob_fd):
        return self.runInteraction(self._write_blob, table, column, irow,
                                   blob_fd)

    def _write_blob(self, trans, table, column, irow, blob_fd):
        blob_fd.seek(0)
        with trans._connection.blob(table, column, irow, 1) as handle:
            data = blob_fd.read(2**12)
            while data:
                handle.write(data)
                data = blob_fd.read(2**12)

    def _blob(self, trans, table, column, irow, flags):
        # TODO: should not use transaction private variable here
        handle = trans._connection.blob(table, column, irow, flags)
        return handle

    def _runInteraction(self, interaction, *args, **kwargs):
        # interaction timeouts are expected for slower machines and lengthy
        # operations because we usually launch several concurrent connections
        # to the sqlcipher backend. The construction below will retry the
        # interaction a number of times before failing in the case of locked
        # database.
        retries = self.timeout_retries
        while True:
            retries -= 1
            try:
                method = adbapi.ConnectionPool._runInteraction
                return method(self, interaction, *args, **kwargs)
            except dbapi2.OperationalError as e:
                if e.message != 'database is locked':
                    # raise if the exception is not about a locked database
                    raise e
                if not retries:
                    # raise if we had already retried enough times
                    logger.warn('database operation timed out, giving up!')
                    raise e
                # log and try again if we can still retry
                logger.warn('database operation timed out, trying again...')



# -*- coding: utf-8 -*-
# sync.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Synchronization between blobs client/server
"""
from collections import defaultdict
from twisted.internet import defer
from twisted.internet import reactor
from twisted.logger import Logger
from twisted.internet import error
from .sql import SyncStatus
from .errors import RetriableTransferError


logger = Logger()


def sleep(seconds):
    d = defer.Deferred()
    reactor.callLater(seconds, d.callback, None)
    return d


MAX_WAIT = 60  # In seconds. Max time between retries


@defer.inlineCallbacks
def with_retry(func, *args, **kwargs):
    """
    Run func repeatedly until success, as long as the exception raised is
    a "retriable error". If an exception of another kind is raised by func,
    the retrying stops and that exception is propagated up the stack.
    """
    retry_wait = 1
    retriable_errors = (error.ConnectError, error.ConnectionClosed,
                        RetriableTransferError,)
    while True:
        try:
            yield func(*args, **kwargs)
            break
        except retriable_errors:
            yield sleep(retry_wait)
            retry_wait = min(retry_wait + 10, MAX_WAIT)


class BlobsSynchronizer(object):

    def __init__(self):
        self.locks = defaultdict(defer.DeferredLock)

    @defer.inlineCallbacks
    def refresh_sync_status_from_server(self, namespace=''):
        d1 = self.remote_list(namespace=namespace)
        d2 = self.local_list(namespace=namespace)
        remote_list, local_list = yield defer.gatherResults([d1, d2])
        pending_download_ids = tuple(set(remote_list) - set(local_list))
        pending_upload_ids = tuple(set(local_list) - set(remote_list))
        yield self.local.update_batch_sync_status(
            pending_download_ids,
            SyncStatus.PENDING_DOWNLOAD,
            namespace=namespace)
        yield self.local.update_batch_sync_status(
            pending_upload_ids,
            SyncStatus.PENDING_UPLOAD,
            namespace=namespace)

    @defer.inlineCallbacks
    def _apply_deletions_from_server(self, namespace=''):
        remote_deletions = self.remote_list(namespace=namespace, deleted=True)
        remote_deletions = yield remote_deletions
        yield self.local.batch_delete(remote_deletions)
        yield self.local.update_batch_sync_status(
            remote_deletions,
            SyncStatus.SYNCED,
            namespace=namespace)

    def send_missing(self, namespace=''):
        """
        Compare local and remote blobs and send what's missing in server.

        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str

        :return: A deferred that fires when all local blobs were sent to
                 server.
        :rtype: twisted.internet.defer.Deferred
        """
        lock = self.locks['send_missing']
        d = lock.run(self._send_missing, namespace)
        return d

    @defer.inlineCallbacks
    def _send_missing(self, namespace):
        # the list of priorities must be refreshed every time a new blob will
        # be transferred. To do that, we use a semaphore and get a new ordered
        # list only when there are free slots for new transfers.
        max_transfers = self.concurrent_transfers_limit
        semaphore = defer.DeferredSemaphore(max_transfers)
        scheduled = set()
        while True:
            d = semaphore.run(self._send_next, namespace, scheduled)
            success = yield d
            if not success:
                break

    @defer.inlineCallbacks
    def _send_next(self, namespace, scheduled):
        status = SyncStatus.PENDING_UPLOAD
        pending = yield self.local_list_status(status, namespace)
        pending = [x for x in pending if x not in scheduled]
        logger.info("There are %d pending blob uploads." % len(pending))

        if not pending:
            # we are finished, indicate that to our caller
            defer.returnValue(False)

        blob_id = pending[0]
        logger.info("Sending blob: %s" % (blob_id,))
        yield with_retry(self._send, blob_id, namespace)
        defer.returnValue(True)

    def fetch_missing(self, namespace=''):
        """
        Compare local and remote blobs and fetch what's missing in local
        storage.

        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str

        :return: A deferred that fires when all remote blobs were received from
                 server.
        :rtype: twisted.internet.defer.Deferred
        """
        lock = self.locks['fetch_missing']
        d = lock.run(self._fetch_missing, namespace)
        return d

    @defer.inlineCallbacks
    def _fetch_missing(self, namespace=''):
        # the list of priorities must be refreshed every time a new blob will
        # be transferred. To do that, we use a semaphore and get a new ordered
        # list only when there are free slots for new transfers.
        max_transfers = self.concurrent_transfers_limit
        semaphore = defer.DeferredSemaphore(max_transfers)
        scheduled = set()
        while True:
            d = semaphore.run(self._fetch_next, namespace, scheduled)
            success = yield d
            if not success:
                break

    @defer.inlineCallbacks
    def _fetch_next(self, namespace, scheduled):
        status = SyncStatus.PENDING_DOWNLOAD
        pending = yield self.local_list_status(status, namespace)
        pending = [x for x in pending if x not in scheduled]
        logger.info("There are %d pending blob downloads." % len(pending))

        if not pending:
            # we are finished, indicate that to our caller
            defer.returnValue(False)

        blob_id = pending[0]
        logger.info("Fetching blob: %s" % (blob_id,))
        yield with_retry(self._fetch, blob_id, namespace)
        defer.returnValue(True)

    @defer.inlineCallbacks
    def sync(self, namespace=''):
        try:
            yield self._apply_deletions_from_server(namespace)
            yield self.refresh_sync_status_from_server(namespace)
            yield self.fetch_missing(namespace)
            yield self.send_missing(namespace)
        except defer.FirstError as e:
            e.subFailure.raiseException()

    @property
    def sync_progress(self):
        return self.local.get_sync_progress()



# -*- coding: utf-8 -*-
# __init__.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Clientside BlobBackend Storage.
"""

from urlparse import urljoin

import os
import json
import base64
import struct

from collections import defaultdict
from io import BytesIO

from twisted.logger import Logger
from twisted.internet import defer

import treq

from leap.soledad.common.errors import SoledadError
from leap.common.files import mkdir_p

from leap.soledad.client._crypto import DocInfo
from leap.soledad.client._crypto import InvalidBlob
from leap.soledad.client._crypto import BlobEncryptor
from leap.soledad.client._crypto import BlobDecryptor
from leap.soledad.client._crypto import EncryptionSchemeNotImplementedException
from leap.soledad.client._crypto import get_unarmored_ciphertext_size
from leap.soledad.client._http import HTTPClient
from leap.soledad.client._pipes import TruncatedTailPipe
from leap.soledad.client._pipes import PreamblePipe

from .sql import SyncStatus
from .sql import Priority
from .sql import SQLiteBlobBackend
from .sync import BlobsSynchronizer
from .upstream_producer import BlobsUpstreamProducer
from .errors import (
    BlobAlreadyExistsError, MaximumRetriesError,
    RetriableTransferError, BlobNotFoundError, InvalidFlagsError)


logger = Logger()
FIXED_REV = 'ImmutableRevision'  # Blob content is immutable
SEPARATOR = ' '


def check_http_status(code, blob_id=None, flags=None):
    if code == 404:
        raise BlobNotFoundError(blob_id)
    if code == 409:
        raise BlobAlreadyExistsError(blob_id)
    elif code == 406:
        raise InvalidFlagsError((blob_id, flags))
    elif code != 200:
        raise SoledadError("Server Error: %s" % code)


class DecrypterBuffer(object):

    def __init__(self, blob_id, secret, tag):
        self.doc_info = DocInfo(blob_id, FIXED_REV)
        self.secret = secret
        self.tag = tag
        self.preamble_pipe = PreamblePipe(self._make_decryptor)
        self.decrypter = None

    def _make_decryptor(self, preamble):
        try:
            self.decrypter = BlobDecryptor(
                self.doc_info, preamble,
                secret=self.secret,
                armor=False,
                start_stream=False,
                tag=self.tag)
            return TruncatedTailPipe(self.decrypter, tail_size=len(self.tag))
        except EncryptionSchemeNotImplementedException:
            # If we do not support the provided encryption scheme, then that's
            # something for the application using soledad to handle. This is
            # the case on asymmetrically encrypted documents on IncomingBox.
            self.raw_data = BytesIO()
            return self.raw_data

    def write(self, data):
        self.preamble_pipe.write(data)

    def close(self):
        if self.decrypter:
            real_size = self.decrypter.decrypted_content_size
            return self.decrypter.endStream(), real_size
        elif hasattr(self, 'raw_data'):
            # blob was not symmetrically encrypted, see _make_decryptor() above
            return self.raw_data, self.raw_data.tell()
        else:
            # the only case that can lead here is if the preamble was not
            # completelly consumed by the pipe.
            msg = "Incomplete Blob: %s" % self.doc_info.doc_id
            raise RetriableTransferError(msg)


class StreamDecrypterBuffer(object):
    size_pack = struct.Struct('<I')

    def __init__(self, secret, blobs_list, done_callback):
        self.blobs_list = blobs_list
        self.secret = secret
        self.done_callback = done_callback
        # self.buf is used to collect size and tag, before becoming a
        # DecrypterBuffer, which then gets used to process the content.
        self.buf = b''
        self.reset()

    def reset(self):
        self.current_blob_size = None
        self.current_blob_id = None
        self.received = 0

    def write(self, data):
        if not self.current_blob_size:
            self.buf += data
            if SEPARATOR in self.buf:
                marker, self.buf = self.buf.split(' ')
                size, tag = marker[:8], marker[8:]
                tag = base64.urlsafe_b64decode(tag)
                self.current_blob_size = int(size, 16)
                self.received = len(self.buf)
                blob_id = self.blobs_list.pop(0)
                buf = DecrypterBuffer(blob_id, self.secret, tag)
                self.current_blob_id = blob_id
                buf.write(self.buf)
                self.buf = buf
        elif (self.received + len(data)) < self.current_blob_size:
            self.buf.write(data)
            self.received += len(data)
        else:
            missing = self.current_blob_size - self.received
            self.buf.write(data[:missing])
            blob_id = self.current_blob_id
            fd, size = self.buf.close()
            self.done_callback(blob_id, fd, size)
            self.buf = data[missing:]
            self.reset()

    def close(self):
        if self.received != 0:
            missing = self.current_blob_size - self.received
            raise Exception("Incomplete download! missing: %s" % missing)
        if self.blobs_list:
            raise Exception("Missing from stream: %s" % self.blobs_list)


class BlobManager(BlobsSynchronizer):
    """
    The BlobManager can list, put, get, set flags and synchronize blobs stored
    in local and remote storages.
    """
    max_decrypt_retries = 3
    concurrent_transfers_limit = 3
    concurrent_writes_limit = 100

    def __init__(
            self, local_path, remote, key, secret, user, token=None,
            cert_file=None, remote_stream=None):
        """
        Initialize the blob manager.

        :param local_path: The path for the local blobs database.
        :type local_path: str
        :param remote: The URL of the remote storage.
        :type remote: str
        :param secret: The secret used to encrypt/decrypt blobs.
        :type secret: str
        :param user: The uuid of the user.
        :type user: str
        :param token: The access token for interacting with remote storage.
        :type token: str
        :param cert_file: The path to the CA certificate file.
        :type cert_file: str
        :param remote_stream: Remote storage stream URL, if supported.
        :type remote_stream: str
        """
        super(BlobsSynchronizer, self).__init__()
        if local_path:
            mkdir_p(os.path.dirname(local_path))
            self.local = SQLiteBlobBackend(local_path, key=key, user=user)
        self.remote = remote
        self.remote_stream = remote_stream
        self.secret = secret
        self.user = user
        self._client = HTTPClient(user, token, cert_file)
        self.semaphore = defer.DeferredSemaphore(self.concurrent_writes_limit)
        self.locks = defaultdict(defer.DeferredLock)

    def close(self):
        if hasattr(self, 'local') and self.local:
            return self.local.close()

    def count(self, namespace=''):
        """
        Count the number of blobs.

        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str
        :return: A deferred that fires with a dict parsed from the JSON
            response, which `count` key has the number of blobs as value.
            Eg.: {"count": 42}
        :rtype: twisted.internet.defer.Deferred
        """
        return self.remote_list(namespace=namespace, only_count=True)

    @defer.inlineCallbacks
    def remote_list(self, namespace='', order_by=None, deleted=False,
                    filter_flag=False, only_count=False):
        """
        List blobs from server, with filtering and ordering capabilities.

        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str
        :param order_by:
            Optional parameter to order results. Possible values are:
            date or +date - Ascending order (older first)
            -date - Descending order (newer first)
        :type order_by: str
        :param deleted:
            Optional paramter to return only deleted blobs.
        :type only_count: bool
        :param filter_flag:
            Optional parameter to filter listing to results containing the
            specified tag.
        :type filter_flag: leap.soledad.common.blobs.Flags
        :param only_count:
            Optional paramter to return only the number of blobs found.
        :type only_count: bool
        :return: A deferred that fires with a list parsed from the JSON
            response, holding the requested list of blobs.
            Eg.: ['blob_id1', 'blob_id2']
        :rtype: twisted.internet.defer.Deferred
        """
        uri = urljoin(self.remote, self.user + '/')
        params = {'namespace': namespace} if namespace else {}
        if order_by:
            params['order_by'] = order_by
        if deleted:
            params['deleted'] = deleted
        if filter_flag:
            params['filter_flag'] = filter_flag
        if only_count:
            params['only_count'] = only_count
        response = yield self._client.get(uri, params=params)
        check_http_status(response.code)
        defer.returnValue((yield response.json()))

    def local_list(self, namespace=''):
        return self.local.list(namespace)

    def local_list_status(self, status, namespace=''):
        return self.local.list_status(status, namespace)

    def put(self, doc, size, namespace='', local_only=False,
            priority=Priority.DEFAULT):
        """
        Put a blob in local storage and upload it to server.

        :param doc: A BlobDoc representing the blob.
        :type doc: leap.soledad.client._document.BlobDoc
        :param size: The size of the blob.
        :type size: int
        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str
        :param local_only: Avoids sync (doesn't send to server).
        :type local_only: bool
        :param priority: Priority for blob upload (one of:
                         low, medium, high, urgent)
        :type priority: str

        :return: A deferred that fires when the blob has been put.
        :rtype: twisted.internet.defer.Deferred
        """
        prio = _parse_priority(priority)
        return self.semaphore.run(
            self._put, doc, size, namespace, local_only, prio)

    @defer.inlineCallbacks
    def _put(self, doc, size, namespace, local_only, priority):
        if (yield self.local.exists(doc.blob_id, namespace=namespace)):
            error_message = "Blob already exists: %s" % doc.blob_id
            raise BlobAlreadyExistsError(error_message)
        fd = doc.blob_fd
        # TODO this is a tee really, but ok... could do db and upload
        # concurrently. not sure if we'd gain something.
        yield self.local.put(doc.blob_id, fd, size=size, namespace=namespace)

        if local_only:
            yield self.local.update_sync_status(
                doc.blob_id, SyncStatus.LOCAL_ONLY, namespace=namespace)
            defer.returnValue(None)

        yield self.local.update_sync_status(
            doc.blob_id, SyncStatus.PENDING_UPLOAD, namespace=namespace,
            priority=priority)
        yield self._send(doc.blob_id, namespace)

    def _send(self, blob_id, namespace):
        lock = self.locks[blob_id]
        d = lock.run(self.__send, blob_id, namespace)
        return d

    @defer.inlineCallbacks
    def __send(self, blob_id, namespace):
        # In fact, some kind of pipe is needed here, where each write on db
        # handle gets forwarded into a write on the connection handle
        fd = yield self.local.get(blob_id, namespace=namespace)
        yield self._encrypt_and_upload(blob_id, fd, namespace=namespace)
        yield self.local.update_sync_status(blob_id, SyncStatus.SYNCED,
                                            namespace=namespace)

    def set_flags(self, blob_id, flags, namespace=''):
        """
        Set flags for a given blob_id.

        :param blob_id:
            Unique identifier of a blob.
        :type blob_id: str
        :param flags:
            List of flags to be set.
        :type flags: [leap.soledad.common.blobs.Flags]
        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str
        :return: A deferred that fires when the operation finishes.
        :rtype: twisted.internet.defer.Deferred
        """
        return self.semaphore.run(self._set_flags, blob_id, flags, namespace)

    @defer.inlineCallbacks
    def _set_flags(self, blob_id, flags, namespace):
        params = {'namespace': namespace} if namespace else None
        flagsfd = BytesIO(json.dumps(flags))
        uri = urljoin(self.remote, self.user + "/" + blob_id)
        response = yield self._client.post(uri, data=flagsfd, params=params)
        check_http_status(response.code, blob_id=blob_id, flags=flags)

    @defer.inlineCallbacks
    def get_flags(self, blob_id, namespace=''):
        """
        Get flags from a given blob_id.

        :param blob_id:
            Unique identifier of a blob.
        :type blob_id: str
        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str
        :return: A deferred that fires with a list parsed from JSON response.
            Eg.: [Flags.PENDING]
        :rtype: twisted.internet.defer.Deferred
        """
        uri = urljoin(self.remote, self.user + "/" + blob_id)
        params = {'namespace': namespace} if namespace else {}
        params['only_flags'] = True
        response = yield self._client.get(uri, params=params)
        check_http_status(response.code, blob_id=blob_id)
        defer.returnValue((yield response.json()))

    @defer.inlineCallbacks
    def get(self, blob_id, namespace='', priority=Priority.DEFAULT):
        """
        Get the blob from local storage or, if not available, from the server.

        :param blob_id:
            Unique identifier of a blob.
        :type blob_id: str
        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str
        :param priority: Priority for blob download (one of:
                         low, medium, high, urgent)
        :type priority: str

        :return: A deferred that fires with the file descriptor for the
                 contents of the blob.
        :rtype: twisted.internet.defer.Deferred
        """
        prio = _parse_priority(priority)
        local_blob = yield self.local.get(blob_id, namespace=namespace)
        if local_blob:
            logger.info("Found blob in local database: %s" % blob_id)
            defer.returnValue(local_blob)

        yield self.local.update_sync_status(
            blob_id, SyncStatus.PENDING_DOWNLOAD, namespace=namespace,
            priority=prio)

        fd = yield self._fetch(blob_id, namespace)
        defer.returnValue(fd)

    def _fetch(self, blob_id, namespace):
        lock = self.locks[blob_id]
        d = lock.run(self.__fetch, blob_id, namespace)
        return d

    @defer.inlineCallbacks
    def __fetch(self, blob_id, namespace):
        try:
            result = yield self._download_and_decrypt(blob_id, namespace)
        except Exception as e:
            _, retries = yield self.local.get_sync_status(blob_id)

            if isinstance(e, InvalidBlob):
                max_retries = self.max_decrypt_retries
                message = "Corrupted blob received from server! ID: %s\n"
                message += "Error: %r\n"
                message += "Retries: %s - Attempts left: %s\n"
                message += "This is either a bug or the contents of the "
                message += "blob have been tampered with. Please, report to "
                message += "your provider's sysadmin and submit a bug report."
                message %= (blob_id, e, retries, (max_retries - retries))
                logger.error(message)

                yield self.local.increment_retries(blob_id)

                if (retries + 1) >= max_retries:
                    failed_download = SyncStatus.FAILED_DOWNLOAD
                    yield self.local.update_sync_status(
                        blob_id, failed_download, namespace=namespace)
                    raise MaximumRetriesError(e)

            raise RetriableTransferError(e)

        if not result:
            defer.returnValue(None)
        blob, size = result

        if blob:
            logger.info("Got decrypted blob of type: %s" % type(blob))
            blob.seek(0)
            yield self.local.put(blob_id, blob, size=size, namespace=namespace)
            yield self.local.update_sync_status(blob_id, SyncStatus.SYNCED,
                                                namespace=namespace)
            local_blob = yield self.local.get(blob_id, namespace=namespace)
            defer.returnValue(local_blob)
        else:
            # XXX we shouldn't get here, but we will...
            # lots of ugly error handling possible:
            # 1. retry, might be network error
            # 2. try later, maybe didn't finished streaming
            # 3.. resignation, might be error while verifying
            logger.error('sorry, dunno what happened')

    @defer.inlineCallbacks
    def _encrypt_and_upload(self, blob_id, fd, namespace=''):
        # TODO ------------------------------------------
        # this is wrong, is doing 2 stages.
        # the crypto producer can be passed to
        # the uploader and react as data is written.
        # try to rewrite as a tube: pass the fd to aes and let aes writer
        # produce data to the treq request fd.
        # ------------------------------------------------
        logger.info("Staring upload of blob: %s" % blob_id)
        doc_info = DocInfo(blob_id, FIXED_REV)
        uri = urljoin(self.remote, self.user + "/" + blob_id)
        crypter = BlobEncryptor(doc_info, fd, secret=self.secret,
                                armor=False)
        fd = yield crypter.encrypt()
        params = {'namespace': namespace} if namespace else None
        response = yield self._client.put(uri, data=fd, params=params)
        check_http_status(response.code, blob_id)
        logger.info("Finished upload: %s" % (blob_id,))

    @defer.inlineCallbacks
    def _downstream(self, blobs_id_list, namespace=''):
        uri = urljoin(self.remote_stream, self.user)
        params = {'namespace': namespace} if namespace else {}
        params['direction'] = 'download'
        data = BytesIO(json.dumps(blobs_id_list))
        response = yield self._client.post(uri, params=params, data=data)
        deferreds = []

        def done_cb(blob_id, blobfd, size):
            d = self.local.put(blob_id, blobfd, size=size, namespace=namespace)
            deferreds.append(d)
        buf = StreamDecrypterBuffer(self.secret, blobs_id_list, done_cb)

        yield treq.collect(response, buf.write)
        yield defer.gatherResults(deferreds, consumeErrors=True)
        buf.close()

    @defer.inlineCallbacks
    def _upstream(self, blobs_id_list, namespace=''):
        local, secret = self.local, self.secret
        uri = urljoin(self.remote_stream, self.user)
        sizes = yield self.local.get_size_list(blobs_id_list, namespace)
        convert = get_unarmored_ciphertext_size
        sizes = map(lambda (blob_id, size): (blob_id, convert(size)), sizes)
        producer = BlobsUpstreamProducer(local, sizes, namespace, secret)
        params = {'namespace': namespace} if namespace else {}
        params['direction'] = 'upload'
        response = yield self._client.post(uri, data=producer, params=params)
        check_http_status(response.code, 'stream')
        logger.info("Finished stream up: %s" % (blobs_id_list,))

    @defer.inlineCallbacks
    def _download_and_decrypt(self, blob_id, namespace=''):
        logger.info("Staring download of blob: %s" % blob_id)
        # TODO this needs to be connected in a tube
        uri = urljoin(self.remote, self.user + '/' + blob_id)
        params = {'namespace': namespace} if namespace else None
        response = yield self._client.get(uri, params=params)
        check_http_status(response.code, blob_id=blob_id)

        if not response.headers.hasHeader('Tag'):
            msg = "Server didn't send a tag header for: %s" % blob_id
            logger.error(msg)
            raise SoledadError(msg)
        tag = response.headers.getRawHeaders('Tag')[0]
        tag = base64.urlsafe_b64decode(tag)
        buf = DecrypterBuffer(blob_id, self.secret, tag)

        # incrementally collect the body of the response
        yield treq.collect(response, buf.write)
        fd, size = buf.close()
        logger.info("Finished download: (%s, %d)" % (blob_id, size))
        defer.returnValue((fd, size))

    def delete(self, blob_id, namespace=''):
        """
        Delete a blob from local and remote storages.

        :param blob_id:
            Unique identifier of a blob.
        :type blob_id: str
        :param namespace:
            Optional parameter to restrict operation to a given namespace.
        :type namespace: str
        :return: A deferred that fires when the operation finishes.
        :rtype: twisted.internet.defer.Deferred
        """
        return self.semaphore.run(self._delete, blob_id, namespace)

    @defer.inlineCallbacks
    def _delete(self, blob_id, namespace):
        logger.info("Marking blobs as PENDING_DELETE: %s" % blob_id)
        yield self.local.update_sync_status(
            blob_id, SyncStatus.PENDING_DELETE, namespace=namespace)
        logger.info("Staring deletion of blob: %s" % blob_id)
        yield self._delete_from_remote(blob_id, namespace=namespace)
        if (yield self.local.exists(blob_id, namespace=namespace)):
            yield self.local.delete(blob_id, namespace=namespace)
        yield self.local.update_sync_status(
            blob_id, SyncStatus.SYNCED, namespace=namespace)

    @defer.inlineCallbacks
    def _delete_from_remote(self, blob_id, namespace=''):
        # TODO this needs to be connected in a tube
        uri = urljoin(self.remote, self.user + '/' + blob_id)
        params = {'namespace': namespace} if namespace else None
        response = yield self._client.delete(uri, params=params)
        check_http_status(response.code, blob_id=blob_id)
        defer.returnValue(response)

    # TODO: evaluate if the following get/set priority methods are needed in
    # the public interface of then blob manager, and remove if not.

    def _set_priority(self, blob_id, priority, namespace=''):
        """
        Set the transfer priority for a certain blob.

        :param blob_id: Unique identifier of a blob.
        :type blob_id: str
        :param priority: The numerical priority to be set.
        :type priority: int
        :param namespace: Optional parameter to restrict operation to a given
                          namespace.
        :type namespace: str

        :return: A deferred that fires after the priority has been set.
        :rtype: twisted.internet.defer.Deferred
        """
        prio = _parse_priority(priority)
        d = self.local.update_priority(blob_id, prio, namespace=namespace)
        return d

    def _get_priority(self, blob_id, namespace=''):
        """
        Get the transfer priority for a certain blob.

        :param blob_id: Unique identifier of a blob.
        :type blob_id: str
        :param namespace: Optional parameter to restrict operation to a given
                          namespace.
        :type namespace: str

        :return: A deferred that fires with the current transfer priority of
                 the blob.
        :rtype: twisted.internet.defer.Deferred
        """
        d = self.local.get_priority(blob_id, namespace=namespace)
        return d


def _parse_priority(prio):
    if isinstance(prio, int):
        if Priority.LOW <= prio <= Priority.URGENT:
            return prio
        else:
            raise ValueError()
    elif isinstance(prio, str):
        if prio == 'low':
            return Priority.LOW
        elif prio == 'medium':
            return Priority.MEDIUM
        elif prio == 'high':
            return Priority.HIGH
        elif prio == 'urgent':
            return Priority.URGENT
        raise ValueError()
    raise ValueError()



# -*- coding: utf-8 -*-
# _blobs/state.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Get handlers that implement IBlobsBackend given a backend name.
"""

from .errors import ImproperlyConfiguredException
from .fs_backend import FilesystemBlobsBackend


class BlobsServerState(object):
    """
    Given a backend name, it gives a instance of IBlobsBackend
    """
    # Allowed backend classes are defined here
    handlers = {"filesystem": FilesystemBlobsBackend}

    def __init__(self, backend, **backend_kwargs):
        if backend not in self.handlers:
            raise ImproperlyConfiguredException("No such backend: %s", backend)
        self.backend = self.handlers[backend](**backend_kwargs)

    def open_database(self, user_id):
        """
        That method is just for compatibility with CouchServerState, so
        IncomingAPI can change backends.
        """
        # TODO: deprecate/refactor it as it's here for compatibility.
        return self.backend



# -*- coding: utf-8 -*-
# incoming.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Code to interact with Incoming Box feature.
See: http://soledad.readthedocs.io/en/latest/incoming_box.html
     or docs/incoming_box.rst
"""
import sys
from leap.soledad.common.blobs import Flags
from twisted.logger import Logger
from twisted.internet.task import LoopingCall
from twisted.internet import defer
log = Logger()


class IncomingBoxProcessingLoop:
    """
    Implements the client-side processing flow for Incoming Box feature,
    maintaining a loop that fetches incoming messages from remote replica,
    delivers it to consumers and changes flags as necessary. This is defined at
    "Processing Incoming Messages" section on Incoming Box Specifications:
    http://soledad.readthedocs.io/en/latest/incoming_box.html
    """

    def __init__(self, incoming_box):
        self.incoming_box = incoming_box
        self.consumers = []
        self._loop = LoopingCall(self._process)

    def __call__(self):
        return self._process()

    @property
    def running(self):
        return self._loop.running

    def start(self, interval=30):
        """
        Starts the inner LoopingCall, triggering the loop.
        :param interval: Time between interactions in seconds.
        :type interval: int
        """
        return self._loop.start(interval)

    def stop(self):
        """
        Stops the inner LoopingCall, stopping the loop.
        """
        return self._loop.stop()

    def add_consumer(self, consumer):
        """
        Adds a consumer to the consumers list, so it can be used to process and
        persist incoming box items.
        :param consumer: Consumer implementation
        :type consumer: leap.soledad.client.interfaces.IIncomingBoxConsumer
        """
        self.consumers.append(consumer)

    @defer.inlineCallbacks
    def _process(self):
        if not self.consumers:
            defer.returnValue(None)
        pending = yield self.incoming_box.list_pending()
        for item_id in pending:
            item = yield self.incoming_box.fetch_for_processing(item_id)
            if not item:
                log.warn("Couldn't reserve item %s for processing, skipping.")
                continue
            failed = False
            for consumer in self.consumers:
                try:
                    parts = yield consumer.process(item, item_id=item_id)
                except Exception:
                    msg = "Consumer %s failed to process item %s: %s"
                    msg %= (consumer.name, item_id, sys.exc_info()[0])
                    log.error(msg)
                    failed = True
                    continue
                yield self.incoming_box.set_processed(item_id)
                try:
                    yield consumer.save(parts, item_id=item_id)
                except Exception:
                    msg = "Consumer %s failed to save item %s: %s"
                    msg %= (consumer.name, item_id, sys.exc_info()[0])
                    log.error(msg)
                    failed = True
            if failed:
                yield self.incoming_box.set_failed(item_id)
            else:
                yield self.incoming_box.delete(item_id)


class IncomingBox:
    """
    A BlobManager proxy that represents an user's Incoming Box.
    It locks all queries to a namespace and deals with parameters and
    implementation details as defined on specification for client/server
    interactions.
    """

    def __init__(self, blob_manager, namespace):
        self.blob_manager = blob_manager
        self.namespace = namespace

    @defer.inlineCallbacks
    def fetch_for_processing(self, blob_id):
        """
        Try to reserve a blob (by flagging it as PROCESSING) and then fetch it.
        If it is already reserved by another replica, return `None` causing the
        loop to skip this item.
        :param blob_id: Unique identifier of a blob.
        :type blob_id: str
        :return: A deferred that fires when operation finishes.
            It will hold None if reservation fails or a file-like object with
            the requested blob.
        :rtype: Deferred
        """
        try:
            yield self.blob_manager.set_flags(blob_id, [Flags.PROCESSING],
                                              namespace=self.namespace)
        except Exception:
            defer.returnValue(None)
        blob = yield self.blob_manager.get(blob_id, namespace=self.namespace)
        defer.returnValue(blob)

    def list_pending(self):
        """
        Lists blobs sorted by date (older first).
        :return: A deferred that fires with the requested list.
        :rtype: Deferred
        """
        return self.blob_manager.remote_list(namespace=self.namespace,
                                             order_by='+date',
                                             filter_flag=Flags.PENDING)

    def set_processed(self, blob_id):
        """
        Flag a blob with Flags.PROCESSED
        :param blob_id: Unique identifier of a blob.
        :type blob_id: str
        :return: A deferred that fires when operation finishes.
        :rtype: Deferred
        """
        return self.blob_manager.set_flags(blob_id, [Flags.PROCESSED],
                                           namespace=self.namespace)

    def set_failed(self, blob_id):
        """
        Flag a blob with Flags.FAILED
        :param blob_id: Unique identifier of a blob.
        :type blob_id: str
        :return: A deferred that fires when operation finishes.
        :rtype: Deferred
        """
        return self.blob_manager.set_flags(blob_id, [Flags.FAILED],
                                           namespace=self.namespace)

    def delete(self, blob_id):
        """
        Deletes a blob belonging to a namespace.
        :param blob_id: Unique identifier of a blob.
        :type blob_id: str
        :return: A deferred that fires when operation finishes.
        :rtype: Deferred
        """
        return self.blob_manager.delete(blob_id, namespace=self.namespace)



# -*- coding: utf-8 -*-
# __init__.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Represents flags that can be used on blobs.
"""
from collections import namedtuple
ACCEPTED_FLAGS = ['PENDING', 'PROCESSING', 'PROCESSED', 'FAILED']
Flags = namedtuple('Flags', ' '.join(ACCEPTED_FLAGS))(*ACCEPTED_FLAGS)


class InvalidFlag(Exception):
    """
    Raised when a flag that is not accepted is used.
    """



# -*- coding: utf-8 -*-
# _http.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
A twisted-based HTTP client that:

    - is pinned to a specific TLS certificate,
    - does token authentication using the Authorization header,
    - can do bandwidth throttling.
"""
import base64
import os
import sys

from twisted.internet import reactor
from twisted.protocols.policies import ThrottlingFactory
from twisted.protocols.policies import ThrottlingProtocol
from twisted.web.iweb import IAgent
from twisted.web.client import Agent as _Agent
from twisted.web.client import CookieAgent
from twisted.web.client import HTTPConnectionPool
from twisted.web.http_headers import Headers

from cookielib import CookieJar
from treq.client import HTTPClient as _HTTPClient
from zope.interface import implementer

from leap.common.http import getPolicyForHTTPS


__all__ = ['HTTPClient']


class HTTPClient(_HTTPClient):

    def __init__(self, uuid, token, cert_file):
        agent = Agent(uuid, token, cert_file)
        jar = CookieJar()
        self._agent = CookieAgent(agent, jar)
        super(self.__class__, self).__init__(self._agent)

    def set_token(self, token):
        self._agent.set_token(token)


class HTTPThrottlingProtocol(ThrottlingProtocol):

    def request(self, *args, **kwargs):
        return self.wrappedProtocol.request(*args, **kwargs)

    def throttleWrites(self):
        if hasattr(self, 'producer') and self.producer:
            self.producer.pauseProducing()

    def unthrottleWrites(self):
        if hasattr(self, 'producer') and self.producer:
            self.producer.resumeProducing()


class HTTPThrottlingFactory(ThrottlingFactory):

    protocol = HTTPThrottlingProtocol


class ThrottlingHTTPConnectionPool(HTTPConnectionPool):

    maxPersistentPerHost = 1          # throttling happens "host-wise"
    maxConnectionCount = sys.maxsize  # max number of concurrent connections
    readLimit = 1 * 10 ** 6           # max bytes we should read per second
    writeLimit = 1 * 10 ** 6          # max bytes we should write per second

    def _newConnection(self, key, endpoint):
        def quiescentCallback(protocol):
            self._putConnection(key, protocol)
        factory = self._factory(quiescentCallback, repr(endpoint))
        throttlingFactory = HTTPThrottlingFactory(
            factory,
            maxConnectionCount=self.maxConnectionCount,
            readLimit=self.readLimit,
            writeLimit=self.writeLimit)
        return endpoint.connect(throttlingFactory)


@implementer(IAgent)
class Agent(_Agent):

    def __init__(self, uuid, token, cert_file, throttling=False):
        self._uuid = uuid
        self._token = None
        self._creds = None
        self.set_token(token)
        factory = getPolicyForHTTPS(cert_file)
        pool = self._get_pool()
        _Agent.__init__(self, reactor, contextFactory=factory, pool=pool)

    def _get_pool(self):
        throttling = bool(os.environ.get('SOLEDAD_THROTTLING'))
        persistent = bool(os.environ.get('SOLEDAD_HTTP_PERSIST'))
        if throttling:
            klass = ThrottlingHTTPConnectionPool
        else:
            klass = HTTPConnectionPool
        return klass(reactor, persistent=persistent)

    def set_token(self, token):
        self._token = token
        self._creds = self._encoded_creds()

    def _encoded_creds(self):
        creds = '%s:%s' % (self._uuid, self._token)
        encoded = base64.b64encode(creds)
        return 'Token %s' % encoded

    def request(self, method, uri, headers=None, bodyProducer=None):
        # authenticate the request
        headers = headers or Headers()
        headers.addRawHeader('Authorization', self._creds)
        # perform the authenticated request
        return _Agent.request(
            self, method, uri, headers=headers, bodyProducer=bodyProducer)



# -*- coding: utf-8 -*-
# config.py
# Copyright (C) 2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import configparser
import os

from twisted.logger import Logger


__all__ = ['get_config']

logger = Logger()

# make sure to update documentation if this default is changed.
DEFAULT_CONFIG_FILE = '/etc/soledad/soledad-server.conf'
CONFIG_DEFAULTS = {
    'soledad-server': {
        'couch_url': 'http://localhost:5984',
        'create_cmd': None,
        'admin_netrc': '/etc/couchdb/couchdb-admin.netrc',
        'batching': True,
        'blobs': False,
        'blobs_path': '/var/lib/soledad/blobs',
        'services_tokens_file': '/etc/soledad/services.tokens',
        'concurrent_blob_writes': 50,
    },
    'database-security': {
        'members': ['soledad'],
        'members_roles': [],
        'admins': [],
        'admins_roles': []
    }
}


def _load_from_file(file_path):
    logger.info('Loading configuration from %s' % file_path)
    conf = dict(CONFIG_DEFAULTS)
    parsed = configparser.SafeConfigParser()
    parsed.read(file_path)
    for section in conf:
        if not parsed.has_section(section):
            continue
        for key, value in conf[section].items():
            if not parsed.has_option(section, key):
                continue
            elif type(value) == bool:
                conf[section][key] = parsed.getboolean(section, key)
            elif type(value) == list:
                values = parsed.get(section, key).split(',')
                values = [v.strip() for v in values]
                conf[section][key] = values
            else:
                conf[section][key] = parsed.get(section, key)
    # TODO: implement basic parsing/sanitization of options comming from
    # parsed file.
    return conf


def _reflect_environment(conf):
    from_environment = ['couch_url']
    for option in from_environment:
        name = 'SOLEDAD_%s' % option.upper()
        value = os.environ.get(name)
        if value:
            logger.info('Using %s=%s because of %s environment variable.'
                        % (option, value, name))
            conf['soledad-server'][option] = value
    return conf


def _load_config(file_path):
    conf = _load_from_file(file_path)
    conf = _reflect_environment(conf)
    return conf


_config = None


def get_config(section='soledad-server'):
    global _config
    if not _config:
        fname = os.environ.get(
            'SOLEDAD_SERVER_CONFIG_FILE', DEFAULT_CONFIG_FILE)
        _config = _load_config(fname)
    return _config[section]



# -*- coding: utf-8 -*-
# state.py
# Copyright (C) 2015,2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Server state using CouchDatabase as backend.
"""
import re

from six.moves.urllib.parse import urljoin

from leap.soledad.common.log import getLogger
from leap.soledad.common.couch import CouchDatabase
from leap.soledad.common.command import exec_validated_cmd
from leap.soledad.common.l2db.remote.server_state import ServerState
from leap.soledad.common.l2db.errors import Unauthorized


logger = getLogger(__name__)


#
# CouchDB Server state
#

def is_db_name_valid(name):
    """
    Validate a user database using a regular expression.

    :param name: database name.
    :type name: str

    :return: boolean for name vailidity
    :rtype: bool
    """
    db_name_regex = "^user-[a-f0-9]+$"
    return re.match(db_name_regex, name) is not None


class CouchServerState(ServerState):

    """
    Inteface of the WSGI server with the CouchDB backend.
    """

    def __init__(self, couch_url, create_cmd=None):
        """
        Initialize the couch server state.

        :param couch_url: The URL for the couch database.
        :type couch_url: str
        :param create_cmd: Command to be executed for user db creation. It will
                           receive a properly sanitized parameter with user db
                           name and should access CouchDB with necessary
                           privileges, which server lacks for security reasons.
        :type create_cmd: str
        """
        self.couch_url = couch_url
        self.create_cmd = create_cmd

    def open_database(self, dbname):
        """
        Open a couch database.

        :param dbname: The name of the database to open.
        :type dbname: str

        :return: The SoledadBackend object.
        :rtype: SoledadBackend
        """
        url = urljoin(self.couch_url, dbname)
        db = CouchDatabase.open_database(url, create=False)
        return db

    def ensure_database(self, dbname):
        """
        Ensure couch database exists.

        :param dbname: The name of the database to ensure.
        :type dbname: str

        :raise Unauthorized: If disabled or other error was raised.

        :return: The SoledadBackend object and its replica_uid.
        :rtype: (SoledadBackend, str)
        """
        if not self.create_cmd:
            raise Unauthorized()
        else:
            code, out = exec_validated_cmd(self.create_cmd, dbname,
                                           validator=is_db_name_valid)
            if code is not 0:
                logger.error("""
                    Error while creating database (%s) with (%s) command.
                    Output: %s
                    Exit code: %d
                    """ % (dbname, self.create_cmd, out, code))
                raise Unauthorized()
        db = self.open_database(dbname)
        return db, db.replica_uid

    def delete_database(self, dbname):
        """
        Delete couch database.

        :param dbname: The name of the database to delete.
        :type dbname: str

        :raise Unauthorized: Always, because Soledad server is not allowed to
                             delete databases.
        """
        raise Unauthorized()



# -*- coding: utf-8 -*-
# interfaces.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from zope.interface import Interface


class IBlobsBackend(Interface):

    """
    An interface for a backend that can store blobs.

    There might be concurrent calls to methods that modify the same blob, so
    it's the backend implementation's responsibility to ensure isolation of
    such actions.
    """

    def read_blob(user, blob_id, consumer, namespace='', range=None):
        """
        Read a blob from the backend storage.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param blob_id: The id of the blob.
        :type blob_id: str
        :param consumer: The object to write data to.
        :type consumer: twisted.internet.interfaces.IConsumer provider
        :param namespace: An optional namespace for the blob.
        :type namespace: str
        :param range: An optional tuple indicating start and end position of
            the blob to be produced.
        :type range: (int, int)

        :return: A deferred that fires when the blob has been written to the
            consumer.
        :rtype: twisted.internet.defer.Deferred

        :raise BlobNotFound: Raised (asynchronously) when the blob was not
            found in the backend.
        """

    def write_blob(user, blob_id, producer, namespace=''):
        """
        Write a blob to the backend storage.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param blob_id: The id of the blob.
        :type blob_id: str
        :param producer: The object to read data from.
        :type producer: twisted.internet.interfaces.IProducer provider
        :param namespace: An optional namespace for the blob.
        :type namespace: str

        :return: A deferred that fires when the blob has been written to the
                 backend storage.
        :rtype: twisted.internet.defer.Deferred

        :raise BlobExists: Raised (asynchronously) when a blob with that id
            already exists.

        :raise QuotaExceeded: Raised (asynchronously) when the quota for that
            user was exceeded.
        """

    def delete_blob(user, blob_id, namespace=''):
        """
        Delete a blob from the backend storage.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param blob_id: The id of the blob.
        :type blob_id: str
        :param namespace: An optional namespace for the blob.
        :type namespace: str

        :return: A deferred that fires when the blob has been deleted.
        :rtype: twisted.internet.defer.Deferred

        :raise BlobNotFound: Raised (asynchronously) when the blob was not
            found in the backend.
        """

    def get_blob_size(user, blob_id, namespace=''):
        """
        Get the size of a blob.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param blob_id: The id of the blob.
        :type blob_id: str
        :param namespace: An optional namespace for the blob.
        :type namespace: str

        :return: A deferred that fires with the size of the blob.
        :rtype: twisted.internet.defer.Deferred

        :raise BlobNotFound: Raised (asynchronously) when the blob was not
            found in the backend.
        """

    def count(user, namespace=''):
        """
        Count the total number of blobs.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param namespace: Restrict the count to a certain namespace.
        :type namespace: str

        :return: A deferred that fires with the number of blobs in the backend
            storage, possibly restricted to a certain namespace.
        :rtype: twisted.internet.defer.Deferred
        """

    def list_blobs(user, namespace='', order_by=None, deleted=False,
                   filter_flag=None):
        """
        List the blobs stored in the backend.

        The resulting list can be ordered by date, filtered by namespace or
        flag, and include deleted items or not.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param namespace: Restrict the count to a certain namespace.
        :type namespace: str
        :param order_by: 'date' (equivalent to '+date') or  '-date', to sort
            ascending or descending by date, respectivelly.
        :type order_by: str
        :param deleted: Whether to include deleted items in the result.
        :type deleted: bool
        :param filter_flag: If given, only results flagged with that flag will
            be returned.
        :type filter_flag: str

        :return: A list of blob ids, optionally ordered and/or restricted by
                 namespace.
        :rtype: list of str

        :return: A deferred that fires with a list of blob ids, optionally
            ordered and/or restricted by namespace.
        :rtype: twisted.internet.defer.Deferred
        """

    def get_total_storage(user):
        """
        Get the size used by a given user as the sum of all the blobs stored
        under all that user's namespaces.

        :param user: The id of a user.
        :type user: str

        :return: The size in units of 1024 bytes.
        :rtype: int

        :return: A deferred that fires with the amount of storage used in units
            of 1024 bytes.
        :rtype: twisted.internet.defer.Deferred
        """

    def get_tag(user, blob_id, namespace=''):
        """
        Get the tag of a blob.

        :param blob_id: The id of the blob.
        :type blob_id: str
        :param namespace: An optional namespace for the blob.
        :type namespace: str

        :return: A deferred that fires with the tag of the blob.
        :rtype: twisted.internet.defer.Deferred

        :raise BlobNotFound: Raised (asynchronously) when the blob was not
            found in the backend.
        """

    def get_flags(user, blob_id, namespace=''):
        """
        Get the flags for a blob.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param blob_id: The id of the blob.
        :type blob_id: str
        :param namespace: An optional namespace for the blob.
        :type namespace: str

        :return: A deferred that fires with the list of flags for a blob.
        :rtype: twisted.internet.defer.Deferred

        :raise BlobNotFound: Raised (asynchronously) when the blob was not
            found in the backend.
        """

    def set_flags(user, blob_id, flags, namespace=''):
        """
        Set flags for a blob.

        :param user: The id of the user who owns the blob.
        :type user: str
        :param blob_id: The id of the blob.
        :type blob_id: str
        :param flags: The list of flags to be set.
        :type flags: list of str
        :param namespace: An optional namespace for the blob.
        :type namespace: str

        :return: A deferred that fires when the flags have been set.
        :rtype: twisted.internet.defer.Deferred

        :raise BlobNotFound: Raised (asynchronously) when the blob was not
            found in the backend.

        :raise InvalidFlag: Raised (asynchronously) when one of the flags
            passed is invalid.
        """



# -*- coding: utf-8 -*-
# preamble.py
# Copyright (C) 2017 LEAP Encryption Access Project
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Preamble is a binary packed metadata payload present on encrypted documents. It
holds data about encryption scheme, iv, document id and sync related data.
   MAGIC, -> used to differentiate from other data formats
   ENC_SCHEME, -> cryptographic scheme (symmetric or asymmetric)
   ENC_METHOD, -> cipher used, such as AES-GCM or AES-CTR or GPG
   current_time, -> time.time()
   self.iv, -> initialization vector if any, or 0 when not applicable
   str(self.doc_id), -> document id
   str(self.rev), -> current revision
   self._content_size) -> size, rounded to ceiling
"""
import base64
import warnings
import struct
import time
from collections import namedtuple
PACMAN = struct.Struct('2sbbQ16s255p255pQ')
LEGACY_PACMAN = struct.Struct('2sbbQ16s255p255p')  # DEPRECATED
MAGIC = '\x13\x37'
ENC_SCHEME = namedtuple('SCHEME', 'symkey external')(1, 2)
ENC_METHOD = namedtuple('METHOD', 'aes_256_ctr aes_256_gcm pgp')(1, 2, 3)
PREAMBLE_SIZE = 736  # 552 urlsafe base64 encoded (it's always armored)


class InvalidPreambleException(Exception):
    pass


class Preamble(object):

    def __init__(self, doc_id, rev, scheme, method,
                 timestamp=0, iv='', magic=None, content_size=0):
        self.doc_id = doc_id
        self.rev = rev
        self.scheme = scheme
        self.method = method
        self.iv = iv
        self.timestamp = int(timestamp) or int(time.time())
        self.magic = magic or MAGIC
        self.content_size = int(content_size)

    def encode(self):
        preamble = PACMAN.pack(
            self.magic,
            self.scheme,
            self.method,
            self.timestamp,
            self.iv,
            str(self.doc_id),
            str(self.rev),
            self.content_size)
        return preamble

    def __eq__(self, other):
        # timestamp insn't included on comparison on purpose since it's not
        # part of the document identity (you can have the very same document at
        # two different times, but you can't have any other field changed and
        # still be able to consider it as the same document).
        fields = ['doc_id', 'rev', 'scheme', 'method', 'iv', 'magic',
                  'content_size']
        for field in fields:
            if not hasattr(other, field):
                return False
            if getattr(self, field) != getattr(other, field):
                return False
        return True


def decode_preamble(encoded_preamble, armored=False):
    if armored:
        encoded_preamble = base64.urlsafe_b64decode(encoded_preamble)
    preamble_size = len(encoded_preamble)
    try:
        if preamble_size == LEGACY_PACMAN.size:
            unpacked_data = LEGACY_PACMAN.unpack(encoded_preamble)
            magic, sch, meth, ts, iv, doc_id, rev = unpacked_data
            warnings.warn("Decoding a legacy preamble without size. " +
                          "This will be deprecated in 0.12. Doc was: " +
                          "doc_id: %s rev: %s" % (doc_id, rev), Warning)
            return Preamble(doc_id, rev, sch, meth, ts, iv, magic)
        elif preamble_size == PACMAN.size:
            unpacked_data = PACMAN.unpack(encoded_preamble)
            magic, sch, meth, ts, iv, doc_id, rev, size = unpacked_data
            return Preamble(doc_id, rev, sch, meth, ts, iv, magic, int(size))
        else:
            raise InvalidPreambleException("Unexpected preamble size %d",
                                           preamble_size)
    except struct.error as e:
        raise InvalidPreambleException(e)



# -*- coding: utf-8 -*-
# check.py
# Copyright (C) 2015,2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Database schema version verification
"""

import treq

from six.moves.urllib.parse import urljoin
from twisted.internet import defer
from urlparse import urlsplit

from leap.soledad.common.couch import CONFIG_DOC_ID
from leap.soledad.common.couch import SCHEMA_VERSION
from leap.soledad.common.couch import SCHEMA_VERSION_KEY
from leap.soledad.common.errors import WrongCouchSchemaVersionError
from leap.soledad.common.errors import MissingCouchConfigDocumentError
from leap.soledad.common.log import getLogger


logger = getLogger(__name__)


@defer.inlineCallbacks
def _check_db_schema_version(url, db, auth, agent=None):
    """
    Check if the schema version is up to date for a given database.

    :param url: the server base URL.
    :type url: str
    :param db: the database name.
    :type db: str
    :param auth: a tuple with (username, password) for acessing CouchDB.
    :type auth: tuple(str, str)
    :param agent: an optional agent for doing requests, used in tests.
    :type agent: twisted.web.client.Agent

    :raise MissingCouchConfigDocumentError: raised when a database is not empty
                                            but has no config document in it.

    :raise WrongCouchSchemaVersionError: raised when a config document was
                                         found but the schema version is
                                         different from what is expected.
    """
    # if there are documents, ensure that a config doc exists
    db_url = urljoin(url, '%s/' % db)
    config_doc_url = urljoin(db_url, CONFIG_DOC_ID)
    res = yield treq.get(config_doc_url, auth=auth, agent=agent)

    if res.code != 200 and res.code != 404:
        raise Exception("Unexpected HTTP response code: %d" % res.code)

    elif res.code == 404:
        res = yield treq.get(urljoin(db_url, '_all_docs'), auth=auth,
                             params={'limit': 1}, agent=agent)
        docs = yield res.json()
        if docs['total_rows'] != 0:
            logger.error(
                "Missing couch config document in database %s" % db)
            raise MissingCouchConfigDocumentError(db)

    elif res.code == 200:
        config_doc = yield res.json()
        if SCHEMA_VERSION_KEY not in config_doc:
            logger.error(
                "Database has config document but no schema version: %s" % db)
            raise WrongCouchSchemaVersionError(db)
        if config_doc[SCHEMA_VERSION_KEY] != SCHEMA_VERSION:
            logger.error(
                "Unsupported database schema in database: %s" % db)
            raise WrongCouchSchemaVersionError(db)


@defer.inlineCallbacks
def check_schema_versions(couch_url, agent=None):
    """
    Check that all user databases use the correct couch schema.

    :param couch_url: The URL for the couch database.
    :type couch_url: str
    :param agent: an optional agent for doing requests, used in tests.
    :type agent: twisted.web.client.Agent
    """
    url = urlsplit(couch_url)
    auth = (url.username, url.password) if url.username else None
    url = "%s://%s:%d" % (url.scheme, url.hostname, url.port)
    deferreds = []
    semaphore = defer.DeferredSemaphore(20)
    logger.info('Starting CouchDB schema versions check...')
    dbs = yield list_dbs(url, auth, agent)
    for db in dbs:
        if not db.startswith('user-'):
            continue
        d = semaphore.run(_check_db_schema_version, url, db, auth, agent=agent)
        deferreds.append(d)
    d = defer.gatherResults(deferreds, consumeErrors=True)
    try:
        yield d
        logger.info('Finished CouchDB schema versions check.')
    except Exception as e:
        msg = 'Error checking CouchDB schema versions: %r' % e
        logger.error(msg)
        raise Exception(msg)


@defer.inlineCallbacks
def list_dbs(url, auth, agent):
    try:
        res = yield treq.get(urljoin(url, '_all_dbs'), auth=auth, agent=agent)
        defer.returnValue((yield res.json()))
    except Exception as e:
        logger.error('Error trying to get list of dbs from %s: %r'
                     % (url, e))
        raise e



# -*- coding: utf-8 -*-
# errors.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
This module contains the different errors that can happen when dealing with
blobs.
"""
from leap.soledad.common.errors import SoledadError


class BlobAlreadyExistsError(SoledadError):
    """
    Raised on attempts to put local or remote blobs that already exist in
    storage.
    """


class BlobNotFoundError(SoledadError):
    """
    Raised on attemtps to get remote blobs that do not exist in storage.
    """


class InvalidFlagsError(SoledadError):
    """
    Raised on attempts to set invalid flags for remotelly stored blobs.
    """


class RetriableTransferError(Exception):
    """
    Raised for any blob transfer error that is considered retriable.
    """


class MaximumRetriesError(Exception):
    """
    Raised when the maximum number of transfer retries has been reached.
    """



# -*- coding: utf-8 -*-
# entrypoints.py
# Copyright (C) 2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Entrypoints for the Soledad server.
"""
from twisted.internet import reactor
from twisted.python import threadpool

from leap.soledad.common.log import getLogger

from .auth import localPortal, publicPortal
from .session import SoledadSession


log = getLogger(__name__)


class UsersEntrypoint(SoledadSession):

    def __init__(self):
        pool = threadpool.ThreadPool(name='wsgi')
        reactor.callWhenRunning(pool.start)
        reactor.addSystemEventTrigger('after', 'shutdown', pool.stop)
        portal = publicPortal(sync_pool=pool)
        SoledadSession.__init__(self, portal)


class ServicesEntrypoint(SoledadSession):

    def __init__(self):
        portal = localPortal()
        SoledadSession.__init__(self, portal)



# -*- coding: utf-8 -*-
# _scrypt.py
# Copyright (C) 2017 LEAP Encryption Access Project
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

try:
    from cryptography.hazmat.backends.interfaces import ScryptBackend
    from cryptography.hazmat.backends import default_backend
    from cryptography.exceptions import UnsupportedAlgorithm
    backend = default_backend()
    OPENSSL_HAS_SCRYPT = isinstance(backend, ScryptBackend)
except ImportError:
    OPENSSL_HAS_SCRYPT = False


def _fallback_hash(secret, salt, buflen=32):
    import scrypt
    return scrypt.hash(secret, salt, buflen=buflen)


if OPENSSL_HAS_SCRYPT:
    from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

    def hash(secret, salt, buflen=32):
        try:
            _hash = Scrypt(
                salt, buflen, 16384, 8, 1, backend).derive(secret)
        except UnsupportedAlgorithm:
            _hash = _fallback_hash(secret, salt, buflen)
        return _hash

else:
    hash = _fallback_hash



# -*- coding: utf-8 -*-
# __main__.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import os
from leap.common.files import mkdir_p
from twisted.internet import defer
from . import BlobManager
from leap.soledad.client._document import BlobDoc
from twisted.logger import Logger

logger = Logger()


#
# testing facilities
#
@defer.inlineCallbacks
def testit(reactor):
    # configure logging to stdout
    from twisted.python import log
    import sys
    log.startLogging(sys.stdout)

    # parse command line arguments
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--url', default='http://localhost:9000/')
    parser.add_argument('--path', default='/tmp/blobs')
    parser.add_argument('--secret', default='secret')
    parser.add_argument('--uuid', default='user')
    parser.add_argument('--token', default=None)
    parser.add_argument('--cert-file', default='')

    subparsers = parser.add_subparsers(help='sub-command help', dest='action')

    # parse upload command
    parser_upload = subparsers.add_parser(
        'upload', help='upload blob and bypass local db')
    parser_upload.add_argument('payload')
    parser_upload.add_argument('blob_id')

    # parse download command
    parser_download = subparsers.add_parser(
        'download', help='download blob and bypass local db')
    parser_download.add_argument('blob_id')
    parser_download.add_argument('--output-file', default='/tmp/incoming-file')

    # parse put command
    parser_put = subparsers.add_parser(
        'put', help='put blob in local db and upload')
    parser_put.add_argument('payload')
    parser_put.add_argument('blob_id')

    # parse get command
    parser_get = subparsers.add_parser(
        'get', help='get blob from local db, get if needed')
    parser_get.add_argument('blob_id')

    # parse delete command
    parser_get = subparsers.add_parser(
        'delete', help='delete blob from local and remote db')
    parser_get.add_argument('blob_id')

    # parse list command
    parser_get = subparsers.add_parser(
        'list', help='list local and remote blob ids')

    # parse send_missing command
    parser_get = subparsers.add_parser(
        'send_missing', help='send all pending upload blobs')

    # parse send_missing command
    parser_get = subparsers.add_parser(
        'fetch_missing', help='fetch all new server blobs')

    # parse arguments
    args = parser.parse_args()

    # TODO convert these into proper unittests

    def _manager():
        mkdir_p(os.path.dirname(args.path))
        manager = BlobManager(
            args.path, args.url,
            'A' * 32, args.secret,
            args.uuid, args.token, args.cert_file)
        return manager

    @defer.inlineCallbacks
    def _upload(blob_id, payload):
        logger.info(":: Starting upload only: %s" % str((blob_id, payload)))
        manager = _manager()
        with open(payload, 'r') as fd:
            yield manager._encrypt_and_upload(blob_id, fd)
        logger.info(":: Finished upload only: %s" % str((blob_id, payload)))

    @defer.inlineCallbacks
    def _download(blob_id):
        logger.info(":: Starting download only: %s" % blob_id)
        manager = _manager()
        result = yield manager._download_and_decrypt(blob_id)
        logger.info(":: Result of download: %s" % str(result))
        if result:
            fd, _ = result
            with open(args.output_file, 'w') as f:
                logger.info(":: Writing data to %s" % args.output_file)
                f.write(fd.read())
        logger.info(":: Finished download only: %s" % blob_id)

    @defer.inlineCallbacks
    def _put(blob_id, payload):
        logger.info(":: Starting full put: %s" % blob_id)
        manager = _manager()
        size = os.path.getsize(payload)
        with open(payload) as fd:
            doc = BlobDoc(fd, blob_id)
            result = yield manager.put(doc, size=size)
        logger.info(":: Result of put: %s" % str(result))
        logger.info(":: Finished full put: %s" % blob_id)

    @defer.inlineCallbacks
    def _get(blob_id):
        logger.info(":: Starting full get: %s" % blob_id)
        manager = _manager()
        fd = yield manager.get(blob_id)
        if fd:
            logger.info(":: Result of get: " + fd.getvalue())
        logger.info(":: Finished full get: %s" % blob_id)

    @defer.inlineCallbacks
    def _delete(blob_id):
        logger.info(":: Starting deletion of: %s" % blob_id)
        manager = _manager()
        yield manager.delete(blob_id)
        logger.info(":: Finished deletion of: %s" % blob_id)

    @defer.inlineCallbacks
    def _list():
        logger.info(":: Listing local blobs")
        manager = _manager()
        local_list = yield manager.local_list()
        logger.info(":: Local list: %s" % local_list)
        logger.info(":: Listing remote blobs")
        remote_list = yield manager.remote_list()
        logger.info(":: Remote list: %s" % remote_list)

    @defer.inlineCallbacks
    def _send_missing():
        logger.info(":: Sending local pending upload docs")
        manager = _manager()
        yield manager.send_missing()
        logger.info(":: Finished sending missing docs")

    @defer.inlineCallbacks
    def _fetch_missing():
        logger.info(":: Fetching remote new docs")
        manager = _manager()
        yield manager.fetch_missing()
        logger.info(":: Finished fetching new docs")

    if args.action == 'upload':
        yield _upload(args.blob_id, args.payload)
    elif args.action == 'download':
        yield _download(args.blob_id)
    elif args.action == 'put':
        yield _put(args.blob_id, args.payload)
    elif args.action == 'get':
        yield _get(args.blob_id)
    elif args.action == 'delete':
        yield _delete(args.blob_id)
    elif args.action == 'list':
        yield _list()
    elif args.action == 'send_missing':
        yield _send_missing()
    elif args.action == 'fetch_missing':
        yield _fetch_missing()


if __name__ == '__main__':
    from twisted.internet.task import react
    react(testit)



# -*- coding: utf-8 -*-
# _blobs/fs_backend.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
A backend for blobs that stores in filesystem.
"""
import base64
import json
import os
import time

from collections import defaultdict
from zope.interface import implementer

from twisted.internet import defer
from twisted.internet import utils
from twisted.web.static import NoRangeStaticProducer
from twisted.web.static import SingleRangeStaticProducer

from leap.common.files import mkdir_p
from leap.soledad.common.blobs import ACCEPTED_FLAGS
from leap.soledad.common.blobs import InvalidFlag
from leap.soledad.common.log import getLogger
from leap.soledad.server import interfaces

from .errors import BlobExists
from .errors import BlobNotFound
from .errors import QuotaExceeded
from .util import VALID_STRINGS


logger = getLogger(__name__)


class NoRangeProducer(NoRangeStaticProducer):
    """
    A static file producer that fires a deferred when it's finished.
    """

    def start(self):
        NoRangeStaticProducer.start(self)
        if self.request is None:
            return defer.succeed(None)
        self.deferred = defer.Deferred()
        return self.deferred

    def stopProducing(self):
        NoRangeStaticProducer.stopProducing(self)
        if hasattr(self, 'deferred'):
            self.deferred.callback(None)


class SingleRangeProducer(SingleRangeStaticProducer):
    """
    A static file producer of a single file range that fires a deferred when
    it's finished.
    """

    def start(self):
        SingleRangeStaticProducer.start(self)
        if self.request is None:
            return defer.succeed(None)
        self.deferred = defer.Deferred()
        return self.deferred

    def stopProducing(self):
        SingleRangeStaticProducer.stopProducing(self)
        if hasattr(self, 'deferred'):
            self.deferred.callback(None)


def isolated(path):
    """
    A decorator that isolates execution of the decorated method using a file
    system lock based on the given path. A symlink in ``{path}.lock`` will be
    used to make sure only one isolated method is executed at a time for that
    path.
    """

    def decorator(method):

        def new_method(*args, **kwargs):
            dirname, _ = os.path.split(path)
            mkdir_p(dirname)
            name = path + '.lock'
            # TODO: evaluate the need to replace this for a readers-writer lock
            lock = defer.DeferredFilesystemLock(name)

            def _release(result):
                lock.unlock()
                return result

            d = lock.deferUntilLocked()
            d.addCallback(lambda _: method(*args, **kwargs))
            d.addCallbacks(_release, _release)
            return d

        return new_method

    return decorator


@implementer(interfaces.IBlobsBackend)
class FilesystemBlobsBackend(object):

    USAGE_TIMEOUT = 30

    def __init__(self, blobs_path='/tmp/blobs/', quota=200 * 1024,
                 concurrent_writes=50):
        self.quota = quota
        self.semaphore = defer.DeferredSemaphore(concurrent_writes)
        if not os.path.isdir(blobs_path):
            os.makedirs(blobs_path)
        self.path = blobs_path
        self.usage = defaultdict(lambda: (None, None))
        self.usage_locks = defaultdict(defer.DeferredLock)

    def __touch(self, path):
        open(path, 'a')

    def read_blob(self, user, blob_id, consumer, namespace='', range=None):
        path = self._get_path(user, blob_id, namespace)
        if not os.path.isfile(path):
            return defer.fail(BlobNotFound((user, blob_id)))

        @isolated(path)
        @defer.inlineCallbacks
        def _read_blob():
            logger.info('reading blob: %s - %s@%s'
                        % (user, blob_id, namespace))
            logger.debug('blob path: %s' % path)
            with open(path) as fd:
                if range is None:
                    producer = NoRangeProducer(consumer, fd)
                else:
                    start, end = range
                    offset = start
                    size = end - start
                    args = (consumer, fd, offset, size)
                    producer = SingleRangeProducer(*args)
                yield producer.start()

        return _read_blob()

    def get_flags(self, user, blob_id, namespace=''):
        path = self._get_path(user, blob_id, namespace)
        if not os.path.isfile(path):
            return defer.fail(BlobNotFound((user, blob_id)))
        if not os.path.isfile(path + '.flags'):
            return defer.succeed([])

        @isolated(path)
        def _get_flags():
            try:
                with open(path + '.flags', 'r') as flags_file:
                    flags = json.loads(flags_file.read())
                    return defer.succeed(flags)
            except Exception as e:
                return defer.fail(e)

        return _get_flags()

    def set_flags(self, user, blob_id, flags, namespace=''):
        path = self._get_path(user, blob_id, namespace)
        if not os.path.isfile(path):
            return defer.fail(BlobNotFound((user, blob_id)))

        @isolated(path)
        def _set_flags():
            try:
                for flag in flags:
                    if flag not in ACCEPTED_FLAGS:
                        raise InvalidFlag(flag)
                with open(path + '.flags', 'w') as flags_file:
                    raw_flags = json.dumps(flags)
                    flags_file.write(raw_flags)
                    return defer.succeed(None)
            except Exception as e:
                return defer.fail(e)

        return _set_flags()

    def write_blob(self, user, blob_id, producer, namespace=''):
        path = self._get_path(user, blob_id, namespace)
        if os.path.isfile(path):
            return defer.fail(BlobExists((user, blob_id)))

        @isolated(path)
        @defer.inlineCallbacks
        def _write_blob():
            try:
                # limit the number of concurrent writes to disk
                yield self.semaphore.acquire()

                try:
                    mkdir_p(os.path.split(path)[0])
                except OSError as e:
                    logger.warn(
                        "Got exception trying to create directory: %r" % e)
                used = yield self.get_total_storage(user)
                length = producer.length / 1024.0
                if used + length > self.quota:
                    raise QuotaExceeded
                logger.info('writing blob: %s - %s' % (user, blob_id))
                with open(path, 'wb') as blobfile:
                    yield producer.startProducing(blobfile)
                used += length
                yield self._update_usage(user, used)
            finally:
                self.semaphore.release()

        return _write_blob()

    @defer.inlineCallbacks
    def _update_usage(self, user, used):
        lock = self.usage_locks[user]
        yield lock.acquire()
        try:
            _, timestamp = self.usage[user]
            self.usage[user] = (used, timestamp)
        finally:
            lock.release()

    def delete_blob(self, user, blob_id, namespace=''):
        path = self._get_path(user, blob_id, namespace)
        if not os.path.isfile(path):
            return defer.fail(BlobNotFound((user, blob_id)))

        @isolated(path)
        def _delete_blob():
            self.__touch(path + '.deleted')
            os.unlink(path)
            try:
                os.unlink(path + '.flags')
            except Exception:
                pass
            return defer.succeed(None)

        return _delete_blob()

    def get_blob_size(self, user, blob_id, namespace=''):
        path = self._get_path(user, blob_id, namespace)
        if not os.path.isfile(path):
            return defer.fail(BlobNotFound((user, blob_id)))

        @isolated(path)
        def _get_blob_size():
            size = os.stat(path).st_size
            try:
                return defer.succeed(size)
            except Exception as e:
                return defer.fail(e)

        return _get_blob_size()

    def count(self, user, namespace=''):
        try:
            base_path = self._get_path(user, namespace=namespace)
        except Exception as e:
            return defer.fail(e)
        count = 0
        for _, _, filenames in os.walk(base_path):
            count += len(filter(lambda i: not i.endswith('.flags'), filenames))
        return defer.succeed(count)

    def list_blobs(self, user, namespace='', order_by=None, deleted=False,
                   filter_flag=False):
        namespace = namespace or 'default'
        blob_ids = []
        try:
            base_path = self._get_path(user, namespace=namespace)
        except Exception as e:
            return defer.fail(e)

        def match(name):
            if deleted:
                return name.endswith('.deleted')
            return VALID_STRINGS.match(name)
        for root, dirs, filenames in os.walk(base_path):
            blob_ids += [os.path.join(root, name) for name in filenames
                         if match(name)]
        if order_by in ['date', '+date']:
            blob_ids.sort(key=lambda x: os.path.getmtime(x))
        elif order_by == '-date':
            blob_ids.sort(key=lambda x: os.path.getmtime(x), reverse=True)
        elif order_by:
            exc = Exception("Unsupported order_by parameter: %s" % order_by)
            return defer.fail(exc)
        if filter_flag:
            blob_ids = list(self._filter_flag(blob_ids, filter_flag))
        blob_ids = [os.path.basename(path).replace('.deleted', '')
                    for path in blob_ids]
        return defer.succeed(blob_ids)

    def _filter_flag(self, blob_paths, flag):
        for blob_path in blob_paths:
            flag_path = blob_path + '.flags'
            if not os.path.isfile(flag_path):
                continue
            with open(flag_path, 'r') as flags_file:
                blob_flags = json.loads(flags_file.read())
            if flag in blob_flags:
                yield blob_path

    @defer.inlineCallbacks
    def get_total_storage(self, user):
        lock = self.usage_locks[user]
        yield lock.acquire()
        try:
            used, timestamp = self.usage[user]
            if used is None or time.time() > timestamp + self.USAGE_TIMEOUT:
                path = self._get_path(user)
                used = yield self._get_disk_usage(path)
                self.usage[user] = (used, time.time())
            defer.returnValue(used)
        finally:
            lock.release()

    def get_tag(self, user, blob_id, namespace=''):
        path = self._get_path(user, blob_id, namespace)
        if not os.path.isfile(path):
            return defer.fail(BlobNotFound((user, blob_id)))

        @isolated(path)
        def _get_tag():
            try:
                with open(path) as doc_file:
                    doc_file.seek(-16, 2)
                    tag = base64.urlsafe_b64encode(doc_file.read())
                    return defer.succeed(tag)
            except Exception as e:
                return defer.fail(e)

        return _get_tag()

    @defer.inlineCallbacks
    def _get_disk_usage(self, start_path):
        if not os.path.isdir(start_path):
            defer.returnValue(0)
        cmd = ['/usr/bin/du', '-s', '-c', start_path]
        output = yield utils.getProcessOutput(cmd[0], cmd[1:])
        size = output.split()[0]
        defer.returnValue(int(size))

    def _validate_path(self, desired_path, user, blob_id):
        if not VALID_STRINGS.match(user):
            raise Exception("Invalid characters on user: %s" % user)
        if blob_id and not VALID_STRINGS.match(blob_id):
            raise Exception("Invalid characters on blob_id: %s" % blob_id)
        desired_path = os.path.realpath(desired_path)  # expand path references
        root = os.path.realpath(self.path)
        if not desired_path.startswith(root + os.sep + user):
            err = "User %s tried accessing a invalid path: %s" % (user,
                                                                  desired_path)
            raise Exception(err)
        return desired_path

    @defer.inlineCallbacks
    def exists(self, user, blob_id, namespace):
        path = self._get_path(user, blob_id, namespace)

        @isolated(path)
        @defer.inlineCallbacks
        def _exists():
            defer.returnValue(os.path.isfile(path))

        return _exists()

    def _get_path(self, user, blob_id='', namespace=''):
        parts = [user]
        if blob_id:
            namespace = namespace or 'default'
            parts += self._get_path_parts(blob_id, namespace)
        elif namespace and not blob_id:
            parts += [namespace]  # namespace path
        else:
            pass  # root path
        path = os.path.join(self.path, *parts)
        return self._validate_path(path, user, blob_id)

    def _get_path_parts(self, blob_id, custom):
        if custom and not blob_id:
            return [custom]
        return [custom] + [blob_id[0], blob_id[0:3], blob_id[0:6]] + [blob_id]



# -*- coding: utf-8 -*-
# _incoming.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
A twisted resource that saves externally delivered documents into user's db.
"""
import base64

from io import BytesIO
from twisted.web.server import NOT_DONE_YET
from twisted.web.resource import Resource
from twisted.web.client import FileBodyProducer

from leap.soledad.common.blobs import Flags
from leap.soledad.common.blobs import preamble
from leap.soledad.common.couch.state import CouchServerState
from leap.soledad.common.crypto import ENC_JSON_KEY
from leap.soledad.common.crypto import ENC_SCHEME_KEY
from leap.soledad.common.crypto import EncryptionSchemes
from leap.soledad.common.document import ServerDocument
from leap.soledad.common.log import getLogger

from leap.soledad.server._config import get_config
from leap.soledad.server._blobs import BlobsServerState
from leap.soledad.server._blobs import BlobExists
from leap.soledad.server._blobs import QuotaExceeded


__all__ = ['IncomingResource']


logger = getLogger(__name__)


def _get_backend_from_config():
    conf = get_config()
    if conf['blobs']:
        return BlobsServerState("filesystem", blobs_path=conf['blobs_path'])
    return CouchServerState(conf['couch_url'])


def uses_legacy(db):
    return hasattr(db, 'put_doc')


class IncomingResource(Resource):
    isLeaf = True

    def __init__(self, backend_factory=None):
        self.factory = backend_factory or _get_backend_from_config()
        self.formatter = IncomingFormatter()

    def render_PUT(self, request):
        uuid, doc_id = request.postpath
        scheme = EncryptionSchemes.PUBKEY
        db = self.factory.open_database(uuid)
        if uses_legacy(db):
            self._put_legacy(doc_id, scheme, db, request)
        else:
            self._put_incoming(uuid, doc_id, scheme, db, request)
        return NOT_DONE_YET

    def _put_legacy(self, doc_id, scheme, db, request):
        try:
            doc = ServerDocument(doc_id)
            content = request.content.read()
            doc.content = self.formatter.format(content, scheme)
            db.put_doc(doc)
            self._finish(request)
        except Exception as e:
            self._error(e, request)

    def _put_incoming(self, user, blob_id, scheme, db, request):
        raw_content = request.content.read()
        preamble = self.formatter.preamble(raw_content, blob_id)
        request.content = BytesIO(preamble + ' ' + raw_content)

        def catchBlobExists(failure):
            failure.trap(BlobExists)
            request.setResponseCode(409)
            request.write("Blob already exists: %s" % blob_id)
            request.finish()

        def catchQuotaExceeded(failure):
            failure.trap(QuotaExceeded)
            logger.error("Error 507: Quota exceeded for user: %s" % user)
            request.setResponseCode(507)
            request.write('Quota Exceeded!')
            request.finish()

        producer = FileBodyProducer(request.content)
        d = db.write_blob(user, blob_id, producer, namespace='MX')
        flags = [Flags.PENDING]
        d.addCallback(lambda _: db.set_flags(user, blob_id, flags,
                                             namespace='MX'))
        d.addCallback(lambda _: request.finish())
        d.addErrback(catchBlobExists)
        d.addErrback(catchQuotaExceeded)
        d.addErrback(self._error, request)

    def _finish(self, request):
        request.write('{"success": true}')
        request.finish()

    def _error(self, e, request):
        logger.error('Error processing request: %s' % e.getErrorMessage())
        request.write('{"success": false}')
        request.setResponseCode(500)
        request.finish()


class IncomingFormatter(object):
    """
    Formats an incoming document. Today as it was by leap_mx and as expected by
    leap_mail, but the general usage should evolve towards a generic way for
    the user to receive external documents.
    """
    INCOMING_KEY = 'incoming'
    ERROR_DECRYPTING_KEY = 'errdecr'  # TODO: Always false on MX, remove it

    def format(self, raw_content, enc_scheme):
        return {self.INCOMING_KEY: True,
                self.ERROR_DECRYPTING_KEY: False,
                ENC_SCHEME_KEY: enc_scheme,
                ENC_JSON_KEY: raw_content}

    def preamble(self, raw_content, doc_id):
        rev = '0'
        scheme = preamble.ENC_SCHEME.external
        method = preamble.ENC_METHOD.pgp
        size = len(raw_content)
        raw_preamble = preamble.Preamble(doc_id, rev, scheme, method,
                                         content_size=size).encode()
        return base64.urlsafe_b64encode(raw_preamble)



# -*- coding: utf-8 -*-
# _blobs/resource.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
A Twisted Web resource for blobs.
"""
import json

from twisted.python.compat import intToBytes
from twisted.python.compat import networkString
from twisted.web import resource
from twisted.web.client import FileBodyProducer
from twisted.web.server import NOT_DONE_YET

from leap.soledad.common.blobs import InvalidFlag
from leap.soledad.server import interfaces

from .fs_backend import FilesystemBlobsBackend
from .errors import BlobNotFound
from .errors import BlobExists
from .errors import ImproperlyConfiguredException
from .errors import QuotaExceeded
from .errors import RangeNotSatisfiable
from .util import VALID_STRINGS

from leap.soledad.common.log import getLogger


logger = getLogger(__name__)


def _catchBlobNotFound(failure, request, user, blob_id):
    failure.trap(BlobNotFound)
    logger.error("Error 404: Blob %s does not exist for user %s"
                 % (blob_id, user))
    request.setResponseCode(404)
    request.write("Blob doesn't exist: %s" % blob_id)
    request.finish()


def _catchBlobExists(failure, request, user, blob_id):
    failure.trap(BlobExists)
    logger.error("Error 409: Blob %s already exists for user %s"
                 % (blob_id, user))
    request.setResponseCode(409)
    request.write("Blob already exists: %s" % blob_id)
    request.finish()


def _catchQuotaExceeded(failure, request, user):
    failure.trap(QuotaExceeded)
    logger.error("Error 507: Quota exceeded for user: %s" % user)
    request.setResponseCode(507)
    request.write('Quota Exceeded!')
    request.finish()


def _catchInvalidFlag(failure, request, user, blob_id):
    failure.trap(InvalidFlag)
    flag = failure.value.message
    logger.error("Error 406: Attempted to set invalid flag %s for blob %s "
                 "for user %s" % (flag, blob_id, user))
    request.setResponseCode(406)
    request.write("Invalid flag: %s" % str(flag))
    request.finish()


def _catchAllErrors(failure, request):
    logger.error('Error processing request: %r' % failure.value)
    request.setResponseCode(500)
    request.finish()


class BlobsResource(resource.Resource):

    isLeaf = True

    # Allowed backend classes are defined here
    handlers = {"filesystem": FilesystemBlobsBackend}

    def __init__(self, backend, blobs_path, **backend_kwargs):
        resource.Resource.__init__(self)
        self._blobs_path = blobs_path
        backend_kwargs.update({'blobs_path': blobs_path})
        if backend not in self.handlers:
            raise ImproperlyConfiguredException("No such backend: %s", backend)
        self._handler = self.handlers[backend](**backend_kwargs)
        assert interfaces.IBlobsBackend.providedBy(self._handler)

    # TODO double check credentials, we can have then
    # under request.

    def _only_count(self, request, user, namespace):
        d = self._handler.count(user, namespace)
        d.addCallback(lambda count: json.dumps({"count": count}))
        d.addCallback(lambda count: request.write(count))
        d.addCallback(lambda _: request.finish())
        return NOT_DONE_YET

    def _list(self, request, user, namespace):
        order = request.args.get('order_by', [None])[0]
        filter_flag = request.args.get('filter_flag', [False])[0]
        deleted = request.args.get('deleted', [False])[0]
        d = self._handler.list_blobs(user, namespace,
                                     order_by=order, deleted=deleted,
                                     filter_flag=filter_flag)
        d.addCallback(lambda blobs: json.dumps(blobs))
        d.addCallback(lambda blobs: request.write(blobs))
        d.addCallback(lambda _: request.finish())
        return NOT_DONE_YET

    def _only_flags(self, request, user, blob_id, namespace):
        d = self._handler.get_flags(user, blob_id, namespace)
        d.addCallback(lambda flags: json.dumps(flags))
        d.addCallback(lambda flags: request.write(flags))
        d.addCallback(lambda _: request.finish())
        d.addErrback(_catchBlobNotFound, request, user, blob_id)
        d.addErrback(_catchAllErrors, request)
        return NOT_DONE_YET

    def _get_blob(self, request, user, blob_id, namespace, range):

        def _set_tag_header(tag):
            request.responseHeaders.setRawHeaders('Tag', [tag])

        def _read_blob(_):
            handler = self._handler
            consumer = request
            d = handler.read_blob(
                user, blob_id, consumer, namespace=namespace, range=range)
            return d

        d = self._handler.get_tag(user, blob_id, namespace)
        d.addCallback(_set_tag_header)
        d.addCallback(_read_blob)
        d.addErrback(_catchBlobNotFound, request, user, blob_id)
        d.addErrback(_catchAllErrors, request)

        return NOT_DONE_YET

    def _parseRange(self, range):
        if not range:
            return None
        try:
            kind, value = range.split(b'=', 1)
            if kind.strip() != b'bytes':
                raise Exception('Unknown unit: %s' % kind)
            start, end = value.split('-')
            start = int(start) if start else None
            end = int(end) if end else None
            return start, end
        except Exception as e:
            raise RangeNotSatisfiable(e)

    def render_GET(self, request):
        logger.info("http get: %s" % request.path)
        user, blob_id, namespace = self._validate(request)
        only_flags = request.args.get('only_flags', [False])[0]

        if not blob_id and request.args.get('only_count', [False])[0]:
            return self._only_count(request, user, namespace)

        if not blob_id:
            return self._list(request, user, namespace)

        if only_flags:
            return self._only_flags(request, user, blob_id, namespace)

        def _handleRangeHeader(size):
            try:
                range = self._parseRange(request.getHeader('Range'))
            except RangeNotSatisfiable:
                content_range = 'bytes */%d' % size
                content_range = networkString(content_range)
                request.setResponseCode(416)
                request.setHeader(b'content-range', content_range)
                request.finish()
                return

            if not range:
                start = end = None
                request.setResponseCode(200)
                request.setHeader(b'content-length', intToBytes(size))
            else:
                start, end = range
                content_range = 'bytes %d-%d/%d' % (start, end, size)
                content_range = networkString(content_range)
                length = intToBytes(end - start)
                request.setResponseCode(206)
                request.setHeader(b'content-range', content_range)
                request.setHeader(b'content-length', length)
            return self._get_blob(request, user, blob_id, namespace, range)

        d = self._handler.get_blob_size(user, blob_id, namespace=namespace)
        d.addCallback(_handleRangeHeader)
        d.addErrback(_catchBlobNotFound, request, user, blob_id)
        d.addErrback(_catchAllErrors, request)
        return NOT_DONE_YET

    def render_DELETE(self, request):
        logger.info("http put: %s" % request.path)
        user, blob_id, namespace = self._validate(request)
        d = self._handler.delete_blob(user, blob_id, namespace=namespace)
        d.addCallback(lambda _: request.finish())
        d.addErrback(_catchBlobNotFound, request, user, blob_id)
        d.addErrback(_catchAllErrors, request)
        return NOT_DONE_YET

    def render_PUT(self, request):
        logger.info("http put: %s" % request.path)
        user, blob_id, namespace = self._validate(request)
        producer = FileBodyProducer(request.content)
        handler = self._handler
        d = handler.write_blob(user, blob_id, producer, namespace=namespace)
        d.addCallback(lambda _: request.finish())
        d.addErrback(_catchBlobExists, request, user, blob_id)
        d.addErrback(_catchQuotaExceeded, request, user)
        d.addErrback(_catchAllErrors, request)
        return NOT_DONE_YET

    def render_POST(self, request):
        logger.info("http post: %s" % request.path)
        user, blob_id, namespace = self._validate(request)
        raw_flags = request.content.read()
        flags = json.loads(raw_flags)
        d = self._handler.set_flags(user, blob_id, flags, namespace=namespace)
        d.addCallback(lambda _: request.write(''))
        d.addCallback(lambda _: request.finish())
        d.addErrback(_catchBlobNotFound, request, user, blob_id)
        d.addErrback(_catchInvalidFlag, request, user, blob_id)
        d.addErrback(_catchAllErrors, request)
        return NOT_DONE_YET

    def _validate(self, request):
        for arg in request.postpath:
            if arg and not VALID_STRINGS.match(arg):
                raise Exception('Invalid blob resource argument: %s' % arg)
        namespace = request.args.get('namespace', ['default'])[0]
        if namespace and not VALID_STRINGS.match(namespace):
            raise Exception('Invalid blob namespace: %s' % namespace)
        return request.postpath + [namespace]



# -*- coding: utf-8 -*-
# _streaming_resource.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
A twisted resource that serves download as a single stream of multiple blobs.
-> POST .../uuid/ DATA: [blob_id, blob_id2, ..., blob_idn]
<- [(size(blob_id), content(blob_id)) for blob_id in DATA] (as a binary stream)
"""
import os
import json
import base64

from zope.interface import implementer
from twisted.internet.interfaces import IPushProducer
from twisted.internet import task, defer, threads
from twisted.web.server import NOT_DONE_YET
from twisted.web.resource import Resource

from leap.common.files import mkdir_p
from leap.soledad.common.log import getLogger
from . import interfaces
from ._blobs import FilesystemBlobsBackend
from ._blobs import ImproperlyConfiguredException


__all__ = ['StreamingResource']


logger = getLogger(__name__)


class StreamingResource(Resource):
    isLeaf = True

    # Allowed backend classes are defined here
    handlers = {"filesystem": FilesystemBlobsBackend}

    def __init__(self, backend, blobs_path, **backend_kwargs):
        Resource.__init__(self)
        self._blobs_path = blobs_path
        backend_kwargs.update({'blobs_path': blobs_path})
        if backend not in self.handlers:
            raise ImproperlyConfiguredException("No such backend: %s", backend)
        self._handler = self.handlers[backend](**backend_kwargs)
        assert interfaces.IBlobsBackend.providedBy(self._handler)

    def render_POST(self, request):
        user = request.postpath[0]
        namespace = request.args.get('namespace', ['default'])[0]
        direction = request.args['direction'][0]
        if direction == 'download':
            return self._startDownstream(user, namespace, request)
        elif direction == 'upload':
            return self._startUpstream(user, namespace, request)
        logger.error("Invalid direction value: %s - %s" % (user, direction))
        request.setResponseCode(500)
        request.write('error, supported direction values are download/upload')
        request.finish()
        return ''

    def _startUpstream(self, user, namespace, request):
        # TODO: at this point, Twisted wrote the request to a temporary file,
        # so it's a disk->disk operation. This has to be improved if benchmark
        # shows its worth.
        args = (user, namespace, request)
        d = threads.deferToThread(self._consume_stream, *args)
        d.addCallback(lambda _: request.finish())

        return NOT_DONE_YET

    def _consume_stream(self, user, namespace, request):
        chunk_size = 2**14
        content = request.content
        incoming_list = json.loads(content.readline())
        for (blob_id, size) in incoming_list:
            db = self._handler
            # TODO: NEEDS SANITIZING
            path = db._get_path(user, blob_id, namespace)
            try:
                mkdir_p(os.path.split(path)[0])
            except OSError as e:
                logger.warn("Got exception trying to create directory: %r" % e)
            with open(path, 'wb') as blob_fd:
                consumed = 0
                while consumed < size:
                    read_size = min(size - consumed, chunk_size)
                    data = content.read(read_size)
                    consumed += read_size
                    blob_fd.write(data)

    def _startDownstream(self, user, namespace, request):
        raw_content = request.content.read()
        blob_ids = json.loads(raw_content)
        deferreds = []
        db = self._handler
        for blob_id in blob_ids:

            def _get_blob_info(blob_id, path):
                d = db.get_blob_size(user, blob_id, namespace)
                d.addCallback(lambda size: (blob_id, path, size))
                return d

            path = db._get_path(user, blob_id, namespace)
            d = _get_blob_info(blob_id, path)
            deferreds.append(d)
        d = defer.gatherResults(deferreds)
        d.addCallback(
            lambda paths: DownstreamProducer(request, paths).start())
        return NOT_DONE_YET


@implementer(IPushProducer)
class DownstreamProducer(object):
    chunk_size = 2**14

    def __init__(self, request, paths):
        self.request = request
        self.paths = paths

    def start(self):
        iterator = self._gen_data()
        self.task = task.cooperate(iterator)
        self.request.registerProducer(self, streaming=True)

    def resumeProducing(self):
        return self.task.resume()

    def pauseProducing(self):
        return self.task.pause()

    def _gen_data(self):
        request, paths = self.request, self.paths
        while paths:
            blob_id, path, size = paths.pop(0)
            request.write('%08x' % size)  # sends file size
            with open(path, 'rb') as blob_fd:
                blob_fd.seek(-16, 2)
                encoded_tag = base64.urlsafe_b64encode(blob_fd.read())
                request.write(encoded_tag)  # sends AES-GCM tag
                blob_fd.seek(0)
                request.write(' ')
                data = blob_fd.read(self.chunk_size)
                while data:
                    yield
                    request.write(data)
                    data = blob_fd.read(self.chunk_size)
        request.unregisterProducer()
        request.finish()

    def stopProducing(self):
        self.request = None
        return self.task.stop()



# -*- coding: utf-8 -*-
# _blobs/__init__.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Blobs Server implementation.
"""
from .fs_backend import FilesystemBlobsBackend
from .resource import BlobsResource
from .state import BlobsServerState
from .errors import BlobExists
from .errors import QuotaExceeded
from .errors import ImproperlyConfiguredException


__all__ = [
    'FilesystemBlobsBackend',
    'BlobsResource',
    'BlobsServerState',
    'BlobExists',
    'QuotaExceeded',
    'ImproperlyConfiguredException',
]


if __name__ == '__main__':
    # A dummy blob server
    # curl -X PUT --data-binary @/tmp/book.pdf localhost:9000/user/someid
    # curl -X GET -o /dev/null localhost:9000/user/somerandomstring
    from twisted.python import log
    import sys
    log.startLogging(sys.stdout)

    from twisted.web.server import Site
    from twisted.internet import reactor

    import argparse

    # parse command line
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', default=9000, type=int)
    parser.add_argument('--path', default='/tmp/blobs/user')
    args = parser.parse_args()

    # run the server
    root = BlobsResource("filesystem", args.path)
    factory = Site(root)
    reactor.listenTCP(args.port, factory)
    reactor.run()



# -*- coding: utf-8 -*-
# session.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Twisted resource containing an authenticated Soledad session.
"""
from zope.interface import implementer

from twisted.cred.credentials import Anonymous
from twisted.cred import error
from twisted.python import log
from twisted.python.components import registerAdapter
from twisted.web import util
from twisted.web._auth import wrapper
from twisted.web.guard import HTTPAuthSessionWrapper
from twisted.web.resource import ErrorPage
from twisted.web.resource import IResource
from twisted.web.server import Session

from zope.interface import Interface
from zope.interface import Attribute

from leap.soledad.server.auth import credentialFactory
from leap.soledad.server.url_mapper import URLMapper


class ISessionData(Interface):
    username = Attribute('An uuid.')
    password = Attribute('A token.')


@implementer(ISessionData)
class SessionData(object):
    def __init__(self, session):
        self.username = None
        self.password = None


registerAdapter(SessionData, Session, ISessionData)


def _sessionData(request):
    session = request.getSession()
    data = ISessionData(session)
    return data


@implementer(IResource)
class UnauthorizedResource(wrapper.UnauthorizedResource):
    isLeaf = True

    def __init__(self):
        pass

    def render(self, request):
        request.setResponseCode(401)
        if request.method == b'HEAD':
            return b''
        return b'Unauthorized'

    def getChildWithDefault(self, path, request):
        return self


@implementer(IResource)
class SoledadSession(HTTPAuthSessionWrapper):

    def __init__(self, portal):
        self._mapper = URLMapper()
        self._portal = portal
        self._credentialFactory = credentialFactory
        # expected by the contract of the parent class
        self._credentialFactories = [credentialFactory]

    def _matchPath(self, request):
        match = self._mapper.match(request.path, request.method)
        return match

    def _parseHeader(self, header):
        elements = header.split(b' ')
        scheme = elements[0].lower()
        if scheme == self._credentialFactory.scheme:
            return (b' '.join(elements[1:]))
        return None

    def _authorizedResource(self, request):
        # check whether the path of the request exists in the app
        match = self._matchPath(request)
        if not match:
            return UnauthorizedResource()

        # get authorization header or fail
        header = request.getHeader(b'authorization')
        if not header:
            return util.DeferredResource(self._login(Anonymous()))

        # parse the authorization header
        auth_data = self._parseHeader(header)
        if not auth_data:
            return UnauthorizedResource()

        # decode the credentials from the parsed header
        try:
            credentials = self._credentialFactory.decode(auth_data, request)
        except error.LoginFailed:
            return UnauthorizedResource()
        except Exception:
            # If you port this to the newer log facility, be aware that
            # the tests rely on the error to be logged.
            log.err(None, "Unexpected failure from credentials factory")
            return ErrorPage(500, None, None)

        # make sure the uuid given in path corresponds to the one given in
        # the credentials
        request_uuid = match.get('uuid')
        if request_uuid and request_uuid != credentials.username:
            return ErrorPage(500, None, None)

        # eventually return a cached resouce
        sessionData = _sessionData(request)
        if sessionData.username == credentials.username \
                and sessionData.password == credentials.password:
            return self._portal.realm.auth_resource

        # if all checks pass, try to login with credentials and cache
        # credentials in case of success
        def _cacheSessionData(res):
            sessionData.username = credentials.username
            sessionData.password = credentials.password
            return res

        d = self._login(credentials)
        d.addCallback(_cacheSessionData)
        return util.DeferredResource(d)



# -*- coding: utf-8 -*-
# app.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Soledad Server, as a Twisted Application.
"""
import os

from twisted.application import service, strports
from twisted.internet import reactor
from twisted.web import server

from leap.soledad.common.couch.check import check_schema_versions
from leap.soledad.common.log import getLogger
from leap.soledad.server import entrypoints
from leap.soledad.server import get_config


logger = getLogger(__name__)


def _deferred_shutdown(status):
    reactor.addSystemEventTrigger('after', 'shutdown',
                                  os._exit, status)
    reactor.stop()


def _exit(status):
    reactor.callWhenRunning(_deferred_shutdown, status)


def _log_and_exit(failure):
    logger.error('Error while starting up server: %r'
                 % failure.getErrorMessage())
    _exit(20)


#
# necessary checks
#

def check_env(local_port, public_port):
    if local_port == public_port:
        logger.error("LOCAL_SERVICES_PORT and HTTPS_PORT can't be the same!")
        _exit(20)

    if public_port is None and not os.getenv('DEBUG_SERVER'):
        logger.error("HTTPS_PORT env var is required to be set!")
        _exit(20)


def check_conf(conf):
    path = conf['blobs_path']
    blobs_not_empty = bool(os.path.exists(path) and os.listdir(path))
    if not conf['blobs'] and blobs_not_empty:
        message = """
**  WARNING: Blobs is disabled, but blobs directory isn't empty.          **
**  If it was previously enabled, disabling can cause data loss due blobs **
**  documents not being accessible to users.                              **
**  Blobs directory: %s
**  REFUSING TO START. Please double check your configuration.            **
    """
        logger.error(message % path)
        _exit(20)


#
# service creation functions
#

def create_local_service(port, application):
    logger.info('Starting local Services HTTP API')
    desc = 'tcp:%s:interface=127.0.0.1' % port
    site = server.Site(entrypoints.ServicesEntrypoint())
    service = strports.service(desc, site)
    service.setServiceParent(application)


def get_tls_service_description(port):
    privateKey = os.getenv('PRIVKEY_PATH', '/etc/soledad/soledad-server.key')
    certKey = os.getenv('CERT_PATH', '/etc/soledad/soledad-server.pem')
    sslmethod = os.getenv('SSL_METHOD', 'SSLv23_METHOD')
    desc = ':'.join([
        'ssl',
        'port=' + str(port),
        'privateKey=' + privateKey,
        'certKey=' + certKey,
        'sslmethod=' + sslmethod])
    return desc


def create_public_service(port, application):
    logger.info('Starting public Users HTTP API')
    if port:
        desc = get_tls_service_description(port)
    else:
        logger.warn('Using plain HTTP on public Users API.')
        desc = 'tcp:port=2424:interface=0.0.0.0'

    site = server.Site(entrypoints.UsersEntrypoint())
    service = strports.service(desc, site)
    service.setServiceParent(application)


def create_services(local_port, public_port, application):
    create_local_service(local_port, application)
    create_public_service(public_port, application)


#
# the application
#

def patch_noisy_factory():
    from twisted.internet.protocol import Factory
    Factory.noisy = False


def run(application):
    patch_noisy_factory()
    local_port = os.getenv('LOCAL_SERVICES_PORT', 2525)
    public_port = os.getenv('HTTPS_PORT', None)
    conf = get_config()
    check_env(local_port, public_port)
    check_conf(conf)
    d = check_schema_versions(conf['couch_url'])
    d.addCallback(lambda _: create_services(local_port, public_port,
                                            application))
    d.addErrback(_log_and_exit)


application = service.Application('soledad-server')



# -*- coding: utf-8 -*-
# _wsgi.py
# Copyright (C) 2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
A WSGI application that serves Soledad synchronization.
"""
from twisted.internet import reactor
from twisted.web.wsgi import WSGIResource

from leap.soledad.server._config import get_config
from leap.soledad.server import SoledadApp
from leap.soledad.server.gzip_middleware import GzipMiddleware
from leap.soledad.common.backend import SoledadBackend
from leap.soledad.common.couch.state import CouchServerState


__all__ = ['get_sync_resource']


def _get_couch_state(conf):
    state = CouchServerState(conf['couch_url'], create_cmd=conf['create_cmd'])
    SoledadBackend.BATCH_SUPPORT = conf.get('batching', False)
    return state


def get_sync_resource(pool):
    conf = get_config()
    state = _get_couch_state(conf)
    app = SoledadApp(state)
    wsgi_app = GzipMiddleware(app)
    return WSGIResource(reactor, pool, wsgi_app)



# -*- coding: utf-8 -*-
# launcher.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Soledad Server launcher.
"""
import argparse
import os
import sys

from twisted.scripts.twistd import run

from leap.soledad import server, __version__


STANDALONE = getattr(sys, 'frozen', False)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='store_true',
                        help='Print the program version.')
    return parser.parse_args()


def here(module=None):
    if STANDALONE:
        # we are running in a |PyInstaller| bundle
        return sys._MEIPASS
    else:
        if module:
            return os.path.dirname(module.__file__)
        else:
            return os.path.dirname(__file__)


def run_server():

    # maybe print version and exit
    args = parse_args()
    if args.version:
        print __version__
        return

    # launch soledad server using twistd
    tac = os.path.join(here(server), 'server.tac')
    args = [
        '--nodaemon',
        '--pidfile=',
        '--syslog',
        '--prefix=soledad-server',
        '--python=%s' % tac,
    ]
    sys.argv[1:] = args
    run()


if __name__ == '__main__':
    run_server()



# -*- coding: utf-8 -*-
# upstream_producer.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import json
from zope.interface import implementer
from twisted.internet import defer
from twisted.internet import reactor
from twisted.web.iweb import IBodyProducer
from twisted.web.iweb import UNKNOWN_LENGTH
from leap.soledad.client._crypto import DocInfo
from leap.soledad.client._crypto import BlobEncryptor


FIXED_REV = 'ImmutableRevision'  # Blob content is immutable


@implementer(IBodyProducer)
class BlobsUpstreamProducer(object):
    """
    Blob producer for upload streams.
    """

    def __init__(self, database, blobs_lengths, namespace, secret):
        """
        Initialize the upload streamer.

        :param database: Local blobs SQLCipher backend instance
        :type database: .sql.SQLiteBlobBackend
        :param blobs_lengths: List of blobs with ids and sizes
        :type blobs_lengths: [(blob_id:str, size:int)]
        :param namespace: Namespace which this stream belongs
        :type namespace: str
        :param secret: The secret used to encrypt blobs.
        :type secret: str
        """
        self.blobs_lengths = blobs_lengths
        self.db = database
        self.length = UNKNOWN_LENGTH
        self.pause = False
        self.stop = False
        self.namespace = namespace
        self.secret = secret

    @defer.inlineCallbacks
    def startProducing(self, consumer):
        """
        Write blobs to the consumer.

        :param consumer: Any IConsumer provider.
        :type consumer: twisted.internet.interfaces.IConsumer

        :return: A Deferred that fires when production ends.
        :rtype: twisted.internet.defer.Deferred
        """
        consumer.write(json.dumps(self.blobs_lengths) + '\n')
        for blob_id, _ in self.blobs_lengths:
            if self.stop:
                break
            if self.pause:
                yield self.sleep(0.001)
                continue
            blob_fd = yield self.db.get(blob_id, namespace=self.namespace)
            doc_info = DocInfo(blob_id, FIXED_REV)
            crypter = BlobEncryptor(doc_info, blob_fd, secret=self.secret,
                                    armor=False)
            fd = yield crypter.encrypt()
            consumer.write(fd.read())

    def sleep(self, secs):
        d = defer.Deferred()
        reactor.callLater(secs, d.callback, None)
        return d

    def pauseProducing(self):
        self.pause = True

    def stopProducing(self):
        self.stop = True

    def resumeProducing(self):
        self.pause = False



# -*- coding: utf-8 -*-
# _blobs/error.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Blobs errors.
"""


class BlobNotFound(Exception):
    """
    Raised when a blob is not found in data storage backend.
    """


class BlobExists(Exception):
    """
    Raised when a blob already exists in data storage backend.
    """


class QuotaExceeded(Exception):
    """
    Raised when the quota would be exceeded if an operation would be held.
    """


class ImproperlyConfiguredException(Exception):
    """
    Raised when there is a problem with the configuration of a backend.
    """


class RangeNotSatisfiable(Exception):
    """
    Raised when the Range: HTTP header was sent but the server doesn't know how
    to satisfy it.
    """



# -*- coding: utf-8 -*-
# setup.py
# Copyright (C) 2013-2017 LEAP Encryption Access Project
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Setup file for leap.soledad
"""
import os
import re
import sys
import versioneer

from setuptools import setup
from setuptools import find_packages
from setuptools.command.develop import develop as _cmd_develop


trove_classifiers = [
    "Intended Audience :: Developers",
    "License :: OSI Approved :: "
    "GNU General Public License v3 or later (GPLv3+)",
    "Environment :: Console",
    "Operating System :: OS Independent",
    "Operating System :: POSIX",
    "Programming Language :: Python :: 2.7",
    "Topic :: Database :: Front-Ends",
    "Topic :: Software Development :: Libraries :: Python Modules"
]

DOWNLOAD_BASE = ('https://github.com/leapcode/soledad/'
                 'archive/%s.tar.gz')
_versions = versioneer.get_versions()
VERSION = _versions['version']
VERSION_REVISION = _versions['full-revisionid']
DOWNLOAD_URL = ""

# get the short version for the download url
_version_short = re.findall('\d+\.\d+\.\d+', VERSION)
if len(_version_short) > 0:
    VERSION_SHORT = _version_short[0]
    DOWNLOAD_URL = DOWNLOAD_BASE % VERSION_SHORT


class cmd_develop(_cmd_develop):
    def run(self):
        # versioneer:
        versions = versioneer.get_versions(verbose=True)
        self._versioneer_generated_versions = versions
        # unless we update this, the command will keep using the old version
        self.distribution.metadata.version = versions["version"]
        _cmd_develop.run(self)


cmdclass = versioneer.get_cmdclass()
cmdclass["develop"] = cmd_develop

install_requires = [
    'pyasn1',
    'service-identity',
    'twisted',
    'treq',
    'paste',   # deprecate
    'routes',  # deprecate
    'six',
    'leap.common',  # deprecate use of asserts
]

client = [
    'cryptography',
    'scrypt',  # XXX deprecate this, cryptography does with recent openssl
    'zope.proxy',
]

# needed until kali merges the py3 fork back into the main pysqlcipher repo
readthedocs = bool(os.environ.get('READTHEDOCS'))
if not readthedocs:
    if sys.version_info.major >= 3:
        client += ['pysqlcipher3']
    else:
        client += ['pysqlcipher']

server = [
    'configparser',
    'Beaker',
    'couchdb'
]

signaling = ['leap.common']


extras = {
    'client': client,
    'server': server,
    'signaling': signaling
}

soledad_server = 'soledad-server=leap.soledad.server.launcher:run_server'

setup(
    name='leap.soledad',
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    url='https://soledad.readthedocs.io/',
    download_url=DOWNLOAD_URL,
    license='GPLv3+',
    description='Synchronization of locally encrypted data among devices.',
    author='The LEAP Encryption Access Project',
    author_email='info@leap.se',
    maintainer='LEAP Encryption Access Project',
    maintainer_email='info@leap.se',
    long_description=(
        "Soledad is the part of LEAP that allows application data to be "
        "securely shared among devices. It provides, to other parts of the "
        "LEAP project, an API for data storage and sync."
    ),
    classifiers=trove_classifiers,
    namespace_packages=["leap"],
    packages=find_packages('src', exclude=['*.tests', '*.tests.*']),
    package_dir={'': 'src'},
    package_data={'': ["*.sql"]},
    install_requires=install_requires,
    extras_require=extras,
    entry_points={'console_scripts': [soledad_server]},
)
